"""Module for the Pearson correlation feature selector.""" ""
from __future__ import annotations

import itertools
from pathlib import Path
from typing import Hashable

import numpy as np
import vaex

from mleko.dataset.data_schema import DataSchema
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr
from mleko.utils.vaex_helpers import get_columns

from .base_feature_selector import BaseFeatureSelector


logger = CustomLogger()
"""A module-level logger for the module."""


class PearsonCorrelationFeatureSelector(BaseFeatureSelector):
    """Selects features based on the Pearson correlation."""

    @auto_repr
    def __init__(
        self,
        correlation_threshold: float,
        features: list[str] | tuple[str, ...] | None = None,
        ignore_features: list[str] | tuple[str, ...] | None = None,
        cache_directory: str | Path = "data/pearson-correlation-feature-selector",
        cache_size: int = 1,
    ) -> None:
        """Initializes the feature selector.

        Will drop one of two features that are highly correlated. The feature to be dropped is the one with the lowest
        average correlation with all other features. If both features have the same average correlation, the first
        feature will be dropped. The default set of features is all numeric features in the DataFrame.

        Note:
            Only works with numeric features.

        Warning:
            Make sure to ignore any important features that need to be kept, such as the
            target feature or some identifier.

        Args:
            correlation_threshold: The maximum correlation allowed for a feature to be selected.
            features: List of feature names to be used by the feature selector.
            ignore_features: List of feature names to be ignored by the feature selector.
            cache_directory: Directory where the cache will be stored locally.
            cache_size: The maximum number of entries to keep in the cache.

        Examples:
            >>> import vaex
            >>> from mleko.dataset.feature_select import PearsonCorrelationFeatureSelector
            >>> from mleko.utils.vaex_helpers import get_column
            >>> df = vaex.from_arrays(
            ...     a=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            ...     b=[1, 2, 3, 4, 5, 6, 7, 8, 9, 9],
            ...     c=[1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            ... )
            >>> ds = DataSchema(numerical=["a", "b", "c"])
            >>> feature_selector = PearsonCorrelationFeatureSelector(
            ...     correlation_threshold=0.75,
            ... )
            >>> ds, _, df = feature_selector.fit_transform(ds, df)
            >>> df.get_column_names()
            ['a', 'c']
        """
        super().__init__(features, ignore_features, cache_directory, cache_size)
        self._correlation_threshold = correlation_threshold
        self._feature_selector: set[str] = set()

    def _fit(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, set[str]]:
        """Fits the feature selector on the input data.

        Args:
            data_schema: The DataSchema of the DataFrame.
            dataframe: The DataFrame to fit the feature selector on.

        Returns:
            Updated DataSchema and the set of features to be dropped.
        """
        features = self._feature_set(data_schema)
        logger.info(f"Fitting pearson correlation feature selector on {len(features)} features: {features}.")

        corr_matrix = abs(np.array(dataframe.correlation(features)))
        avg_corr = corr_matrix.mean(axis=1)

        # Generate all possible pairs of features
        feature_pairs = list(itertools.combinations(enumerate(features), 2))

        # Find correlated feature pairs and the feature to drop based on average correlation
        correlated_features = [
            (f_a, f_b, f_a if avg_corr[idx_a] > avg_corr[idx_b] else f_b)
            for (idx_a, f_a), (idx_b, f_b) in feature_pairs
            if corr_matrix[idx_a, idx_b] >= self._correlation_threshold
        ]

        # Create sets of all correlated features and potentially dropped features
        all_correlated_features = {f_a for f_a, _, _ in correlated_features}.union(
            f_b for _, f_b, _ in correlated_features
        )
        potentially_dropped = {drop_f for _, _, drop_f in correlated_features}

        # Identify features guaranteed to be kept
        guaranteed_kept = all_correlated_features.difference(potentially_dropped)

        # Identify features guaranteed to be dropped
        guaranteed_dropped = {
            f
            for f_a, f_b, _ in correlated_features
            if f_a in guaranteed_kept or f_b in guaranteed_kept
            for f in (f_a, f_b)
            if f not in guaranteed_kept
        }

        # Identify features that are not guaranteed to be dropped but are highly correlated
        possible_drop = potentially_dropped.difference(guaranteed_dropped)

        # Identify additional features to be dropped based on correlation with possible_drop features
        additional_dropped = {
            drop_f
            for f_a, f_b, drop_f in correlated_features
            if (f_a in possible_drop or f_b in possible_drop)
            and (f_a not in guaranteed_dropped and f_b not in guaranteed_dropped)
        }

        # Combine guaranteed_dropped and additional_dropped sets
        self._feature_selector = guaranteed_dropped.union(additional_dropped)
        ds = data_schema.copy().drop_features(self._feature_selector)

        return ds, self._feature_selector

    def _transform(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, vaex.DataFrame]:
        """Selects features based on the Pearson correlation.

        Args:
            data_schema: The DataSchema of the DataFrame.
            dataframe: The DataFrame to select features from.

        Returns:
            The updated DataSchema and DataFrame with the selected features.
        """
        dropped_features = self._feature_selector
        logger.info(
            f"Dropping ({len(dropped_features)}) features with correlation >= {self._correlation_threshold}: "
            f"{dropped_features}."
        )
        selected_features = [feature for feature in dataframe.get_column_names() if feature not in dropped_features]
        ds = data_schema.copy().drop_features(dropped_features)

        return ds, get_columns(dataframe, selected_features)

    def _default_features(self, data_schema: DataSchema) -> tuple[str, ...]:
        """Returns the default set of features.

        Args:
            data_schema: The DataSchema of the DataFrame.

        Returns:
            Tuple of default features.
        """
        features = data_schema.get_features(["numerical", "boolean"])
        return tuple(str(feature) for feature in features)

    def _fingerprint(self) -> Hashable:
        """Returns a hashable fingerprint of the feature selector.

        Append the pearson correlation threshold to the fingerprint.

        Returns:
            The fingerprint of the feature selector.
        """
        return super()._fingerprint(), self._correlation_threshold



"""A feature selector that combines multiple feature selectors."""

from __future__ import annotations

from pathlib import Path
from typing import Any, Hashable

import vaex

from mleko.dataset.data_schema import DataSchema
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr

from .base_feature_selector import BaseFeatureSelector


logger = CustomLogger()
"""A module-level logger for the module."""


class CompositeFeatureSelector(BaseFeatureSelector):
    """A feature selector that combines multiple feature selectors.

    It is possible to combine multiple feature selectors into a single feature selector. This can be useful when
    multiple feature selectors need to be applied to a DataFrame and the cache needs to be shared between them.
    """

    @auto_repr
    def __init__(
        self,
        feature_selectors: list[BaseFeatureSelector] | tuple[BaseFeatureSelector, ...],
        cache_directory: str | Path = "data/composite-feature-selector",
        cache_size: int = 1,
    ) -> None:
        """Initializes the composite feature selector.

        The composite feature selector will combine the feature selectors into a single feature selector. Each feature
        selector will be applied to the DataFrame in the order they are specified.

        Args:
            feature_selectors: List of feature selectors to be combined.
            cache_directory: Directory where the cache will be stored locally.
            cache_size: The maximum number of entries to keep in the cache.

        Examples:
            >>> import vaex
            >>> from mleko.dataset.feature_select import CompositeFeatureSelector, MissingRateFeatureSelector
            >>> df = vaex.from_arrays(
            ...     a=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            ...     b=[1, 2, 3, 4, None, None, None, None, None, None],
            ...     c=[1, 2, 3, 4, 5, 6, None, None, None, None],
            ... )
            >>> ds = DataSchema(
            ...     numerical=["a", "b", "c"],
            ... )
            >>> feature_selector = CompositeFeatureSelector(
            ...     feature_selectors=[
            ...         MissingRateFeatureSelector(
            ...             missing_rate_threshold=0.75,
            ...         ),
            ...         MissingRateFeatureSelector(
            ...             missing_rate_threshold=0.50,
            ...         ),
            ...     ],
            ... )
            >>> ds, _, df = feature_selector.fit_transform(ds, df)
            >>> df
            #    a    c
            0    1    1
            1    2    2
            2    3    3
            3    4    4
            4    5    5
            5    6    6
            6    7    None
            7    8    None
            8    9    None
            9   10    None
        """
        super().__init__(None, None, cache_directory, cache_size)
        self._feature_selectors = tuple(feature_selectors)
        self._feature_selector: list[Any] = []

    def _fit(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, list[Any]]:
        """Fits the feature selector on the DataFrame.

        Args:
            data_schema: DataSchema of the DataFrame.
            dataframe: DataFrame on which the feature selector will be fitted.

        Returns:
            Tuple of updated DataSchema and list of fitted feature selectors.
        """
        feature_selectors: list[Any] = []
        for i, feature_selector in enumerate(self._feature_selectors):
            logger.info(
                f"Fitting composite feature selection step {i+1}/{len(self._feature_selectors)}: "
                f"{feature_selector.__class__.__name__}."
            )
            data_schema, feature_selector = feature_selector.fit(data_schema, dataframe, disable_cache=True)
            feature_selectors.append(feature_selector)
            logger.info(f"Finished fitting composite feature selection step {i+1}/{len(self._feature_selectors)}.")
        return data_schema, feature_selectors

    def _transform(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, vaex.DataFrame]:
        """Selects the features from the DataFrame.

        Args:
            data_schema: DataSchema of the DataFrame.
            dataframe: DataFrame from which the features will be selected.

        Returns:
            DataFrame with the selected features.
        """
        for i, feature_selector in enumerate(self._feature_selectors):
            logger.info(
                f"Executing composite feature selection step {i+1}/{len(self._feature_selectors)}: "
                f"{feature_selector.__class__.__name__}."
            )
            data_schema, dataframe = feature_selector.transform(data_schema, dataframe, disable_cache=True)
            dataframe = dataframe.extract()
            logger.info(f"Finished composite feature selection step {i+1}/{len(self._feature_selectors)}.")
        return data_schema, dataframe

    def _fit_transform(
        self, data_schema: DataSchema, dataframe: vaex.DataFrame
    ) -> tuple[DataSchema, Any, vaex.DataFrame]:
        """Fits the feature selector to the specified DataFrame and extracts the selected features from the DataFrame.

        Args:
            data_schema: DataSchema of the DataFrame.
            dataframe: DataFrame used for feature selection.

        Returns:
            Tuple of updated data schema, fitted feature selector and transformed DataFrame.
        """
        feature_selectors: list[Any] = []
        for i, feature_selector in enumerate(self._feature_selectors):
            logger.info(
                f"Executing composite feature selection step {i+1}/{len(self._feature_selectors)}: "
                f"{feature_selector.__class__.__name__}."
            )
            data_schema, feature_selector, dataframe = feature_selector.fit_transform(
                data_schema, dataframe, disable_cache=True
            )
            feature_selectors.append(feature_selector)
            dataframe = dataframe.extract()
            logger.info(
                "Finished fitting and transforming composite feature "
                f"selection step {i+1}/{len(self._feature_selectors)}."
            )
        return data_schema, feature_selectors, dataframe

    def _assign_feature_selector(self, feature_selector: Any) -> None:
        """Assigns the specified feature selector to the feature_selector attribute.

        Can be overridden by subclasses to assign the feature selector using a different method.

        Args:
            feature_selector: Feature selector to be assigned.
        """
        self._feature_selector = feature_selector
        for feature_selector_obj, fitted_feature_selector in zip(self._feature_selectors, feature_selector):
            feature_selector_obj._feature_selector = fitted_feature_selector

    def _default_features(self, data_schema: DataSchema) -> tuple[str, ...]:  # pragma: no cover
        """Returns the default features of the DataFrame.

        Args:
            data_schema: DataSchema of the DataFrame.

        Returns:
            Tuple of default features.
        """
        features = data_schema.get_features()
        return tuple(str(feature) for feature in features)

    def _fingerprint(self) -> Hashable:
        """Returns the fingerprint of the feature selector.

        Returns:
            Fingerprint of the feature selector.
        """
        return super()._fingerprint(), tuple(
            feature_selector._fingerprint() for feature_selector in self._feature_selectors
        )



"""This module provides utility functions for file and directory operations."""

from __future__ import annotations

from pathlib import Path


def clear_directory(directory: Path, pattern: str = "*") -> None:
    """Remove all files in a directory that match a given pattern.

    This function takes a directory and, using the provided pattern, searches for all matching files
    and removes them. This is useful when cleaning up temporary or intermediate files in a workspace.

    Args:
        directory: The `Path` object referring to the directory to be cleared.
        pattern: The search pattern to match the files in the directory (default: "*", matches all files).
    """
    for f in directory.glob(pattern):
        f.unlink()



"""This module defines the abstract base class for pipeline steps in a data processing pipeline.

The module provides a standard interface for implementing data processing steps as part of a larger pipeline,
via the `PipelineStep` abstract base class. Each `PipelineStep` subclass should have a specific purpose and
should be able to run independently or as part of the pipeline.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Literal, TypeVar

from typing_extensions import TypedDict, get_type_hints

from mleko.pipeline.data_container import DataContainer
from mleko.utils.custom_logger import CustomLogger


logger = CustomLogger()
"""The logger for the module."""

T = TypeVar("T")
"""Type variable for generic type annotations."""

TypedDictType = TypeVar("T", bound=TypedDict)  # type: ignore
"""Type variable for TypedDict type annotations."""


def validate_typeddict(data: dict[str, Any], typeddict: type[TypedDictType]) -> None:
    """Validates that all keys defined in the TypedDict are present in the data dict.

    This function checks for the presence of all keys, including those marked as Optional.
    The presence of a key is mandatory, but the value can be None if the key is Optional.

    Args:
        data: The data dictionary to validate.
        typeddict: The TypedDict class against which to validate the data.

    Raises:
        ValueError: If the data is missing required keys or has extra keys not defined in the TypedDict.
    """
    type_hints = get_type_hints(typeddict)

    # Check for missing keys
    missing_keys = [key for key in type_hints if key not in data]
    if missing_keys:
        msg = f"Missing required key(s) in {typeddict.__name__}: {', '.join(missing_keys)}"
        logger.error(msg)
        raise ValueError(msg)

    # Check for extra keys
    extra_keys = [key for key in data if key not in type_hints]
    if extra_keys:
        msg = f"Extra key(s) in data not defined in {typeddict.__name__}: {', '.join(extra_keys)}"
        logger.error(msg)
        raise ValueError(msg)


class PipelineStep(ABC):
    """Base class for all pipeline steps, ensuring the standardized interface for performing data processing operations.

    Descendants of this class must implement the `execute` method, which carries out the data processing operation
    related to the step. Additionally, they must implement the `_get_input_model` and `_get_output_model` methods,
    which dynamically determine the TypedDict models to use for validation.
    """

    def __init__(
        self,
        inputs: dict[str, Any] | TypedDictType,
        outputs: dict[str, str] | TypedDictType,
        cache_group: str | None,
    ) -> None:
        """Initialize a new PipelineStep with the provided input and output keys.

        Args:
            action: The action to perform, such as "fit", "transform", or "fit_transform".
            inputs: Input data required for this step's processing operation.
            outputs: Output data generated by this step's processing operation.
            cache_group: The cache group to use.
        """
        self._inputs = inputs
        self._outputs = outputs
        self._cache_group = cache_group

        self._validate_inputs()
        self._validate_outputs()

    @abstractmethod
    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Execute the data processing operation associated with this pipeline step.

        The `execute` method is the main entry point for the data processing operation associated with this step.
        It receives a `DataContainer` instance as input, containing the data to be processed by this step.
        The method should perform the processing operation and return the processed data as a `DataContainer` instance.

        Args:
            data_container: Input data for this step's processing operation.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Raises:
            NotImplementedError: Must be implemented by subclass.
        """
        raise NotImplementedError

    @abstractmethod
    def _get_input_model(self) -> type[TypedDictType]:  # type: ignore
        """Dynamically determine the TypedDict model to use for validation.

        Returns:
            The TypedDict model to use for validation.
        """
        pass

    @abstractmethod
    def _get_output_model(self) -> type[TypedDictType]:  # type: ignore
        """Dynamically determine the TypedDict model to use for validation.

        Returns:
            The TypedDict model to use for validation.
        """
        pass

    def _validate_and_get_input(
        self,
        input_object: str | T,
        expected_type: type[T],
        data_container: DataContainer,
        is_optional: bool = False,
    ) -> T:
        """Validate and get the input from the data container or as a direct value.

        If the input is a string, it is treated as a key to look up in the data container. If it is not a string,
        it is treated as the input value itself. The method validates the input type, and returns the input value
        if it is valid.

        Warning:
            This method does not handle subscripted types, such as `List[str]`. It only handles simple types. If you
            need to validate a subscripted type, you should do so manually after retrieving the input value from this
            method.

        Args:
            input_object: The input key or value to validate and retrieve.
            expected_type: The expected type of the input value.
            data_container: The data container containing the input data.
            is_optional: Whether the input is optional.

        Raises:
            ValueError: If the input is invalid or not found in the data container.

        Returns:
            The input value if it is valid, or None if it is optional and not found in the data container.
        """
        if isinstance(input_object, str):
            input_value: T = data_container.data.get(input_object, None)
        else:
            input_value: T = input_object

        if input_value is None and not is_optional and isinstance(input_object, str):
            msg = f"Input {input_object!r} is required but not found in data container."
            logger.error(msg)
            raise ValueError(msg)

        if input_value is None and not is_optional and not isinstance(input_object, str):
            msg = "Input is required but found None."
            logger.error(msg)
            raise ValueError(msg)

        if input_value is not None and not isinstance(input_value, expected_type):
            msg = f"Invalid data type: {type(input_value)}. Expected {expected_type}."
            logger.error(msg)
            raise ValueError(msg)

        return input_value

    def _validate_inputs(self) -> None:
        """Validates the step's inputs using TypeDict models.

        Raises:
            ValueError: If the inputs are not a dictionary.
        """
        if isinstance(self._inputs, dict):
            input_model = self._get_input_model()
            validate_typeddict(self._inputs, input_model)
        else:
            msg = f"Inputs must be a dictionary, got {type(self._inputs)}."
            logger.error(msg)
            raise ValueError(msg)

    def _validate_outputs(self) -> None:
        """Validates the step's outputs using TypeDict models.

        Raises:
            ValueError: If the outputs are not a dictionary.
        """
        if isinstance(self._outputs, dict):
            output_model = self._get_output_model()
            validate_typeddict(self._outputs, output_model)
        else:
            msg = f"Outputs must be a dictionary, got {type(self._outputs)}."
            logger.error(msg)
            raise ValueError(msg)


FitTransformAction = Literal["fit", "transform", "fit_transform"]
"""Type alias for the action to perform, either "fit", "transform", or "fit_transform"."""


class FitTransformPipelineStep(PipelineStep):
    """Base class for all pipeline steps, ensuring the standardized interface for performing data processing operations.

    Specialized for steps that perform actions "fit", "transform", or "fit_transform". Descendants of this class must
    implement the `execute` method, which carries out the data processing operation related to the step. Additionally,
    they must implement the `_get_input_model` and `_get_output_model` methods, which dynamically determine the
    TypedDict models to use for validation based on the action to be performed.
    """

    _action: FitTransformAction

    def __init__(
        self,
        action: FitTransformAction,
        inputs: dict[str, Any] | TypedDictType,
        outputs: dict[str, str] | TypedDictType,
        cache_group: str | None,
    ) -> None:
        """Initialize a new PipelineStep with the provided input and output keys.

        Args:
            action: The action to perform, either "fit", "transform", or "fit_transform".
            inputs: Input data required for this step's processing operation.
            outputs: Output data generated by this step's processing operation.
            cache_group: The cache group to use.

        Raises:
            ValueError: If the action is not one of "fit", "transform", or "fit_transform".
        """
        if action not in ("fit", "transform", "fit_transform"):
            raise ValueError(f"Invalid action: {action}. Expected one of 'fit', 'transform', or 'fit_transform'.")

        self._action = action
        super().__init__(inputs, outputs, cache_group)



"""This module provides utility decorators for classes and functions."""

from __future__ import annotations

import inspect
from functools import wraps
from time import perf_counter
from typing import Any, Callable, TypeVar, cast

from mleko.utils.custom_logger import CustomLogger


logger = CustomLogger()
"""A CustomLogger instance that's used throughout the module for logging."""

F = TypeVar("F", bound=Callable[..., Any])
"""A TypeVar used as a generic function type throughout the module.

This TypeVar is designed for type hinting in decorators in the module. It essentially states that F can be
a function with any number of arguments and any return type.
"""


def auto_repr(init_method: F) -> F:
    """Decorator for generating a `__repr__` method for a class automatically based on the `__init__` method signature.

    The decorator inspects the `__init__` method's signature and uses parameter names and values to create a `__repr__`
    method that represents the class instance.

    Args:
        init_method: The `__init__` method of the class to be decorated.

    Returns:
        The wrapped `__init__` method with an automatically generated `__repr__` method.

    Examples:
        >>> from mleko.utils import auto_repr
        >>> class TestClass:
        ...     @auto_repr
        ...     def __init__(self, param1: int, param2: str = "default"):
        ...         self.param1 = param1
        ...         self.param2 = param2
        >>> test_instance = TestClass(1, "test")
        >>> test_instance
        TestClass(param1=1, param2='test')
    """

    @wraps(init_method)
    def wrapped_init(self: Any, *args: Any, **kwargs: Any) -> None:
        init_method(self, *args, **kwargs)

        signature = inspect.signature(init_method)
        parameter_names = list(signature.parameters.keys())[1:]
        parameter_values = dict(zip(parameter_names, args))

        for key, param in signature.parameters.items():
            if param.default != inspect.Parameter.empty:
                parameter_values.setdefault(key, param.default)

        parameter_values.update(kwargs)

        self._auto_repr_args_kwargs = parameter_values
        self.__class__.__repr__ = generated_repr

    def generated_repr(self: Any) -> str:
        cls_name = type(self).__name__
        params_str = ", ".join(f"{key}={repr(value)}" for key, value in self._auto_repr_args_kwargs.items())
        return f"{cls_name}({params_str})"

    return cast(F, wrapped_init)


def timing(func: F) -> F:
    """A decorator that logs the execution time of the decorated function using a CustomLogger instance.

    Warning:
        On certain platforms, the timing may be inaccurate due to OS scheduling. This is a known issue with some
        MacOS versions.

    Args:
        func: The function to be decorated.

    Returns:
        The wrapped function that logs its execution time.

    Example:
        >>> from mleko.utils import timing
        >>> @timing
        ... def test_func():
        ...     pass
        >>> test_func()
        Function: test_func    Timing: 0.0000s
    """

    @wraps(func)
    def wrap(*args: Any, **kwargs: Any) -> Any:
        ts = perf_counter()
        result = func(*args, **kwargs)
        te = perf_counter()
        elapsed_time = te - ts
        logger.debug(f"Function: {func.__qualname__}    Timing: {elapsed_time:.4f}s")
        return result

    return cast(F, wrap)



"""The module contains a fingerprinter for dictionaries."""

from __future__ import annotations

import hashlib
import json
from typing import Any

from .base_fingerprinter import BaseFingerprinter


class DictFingerprinter(BaseFingerprinter):
    """Class to generate unique fingerprints for dictionaries."""

    def fingerprint(self, data: dict[str, Any] | None) -> str:
        """Generate a fingerprint string for a given dictionary.

        Args:
            data: The dictionary to dump.

        Returns:
            A fingerprint that uniquely identifies the dictionary.
        """

        def deep_sort(obj: dict | Any):
            """Recursively sort nested dicts.

            Args:
                obj: The dict or object to sort.

            Returns:
                Sorted dict, or object.
            """
            if isinstance(obj, dict):
                _sorted = {}
                for key in sorted(obj):
                    _sorted[key] = deep_sort(obj[key])
                return _sorted
            else:
                return obj

        sorted_data = deep_sort(data)
        return hashlib.md5((json.dumps(sorted_data, sort_keys=True) if data is not None else "").encode()).hexdigest()



"""Module for DataSchema class, used for storing type information about the dataset."""

from __future__ import annotations

import copy
from typing import Literal

from mleko.cache.fingerprinters.dict_fingerprinter import DictFingerprinter
from mleko.utils.custom_logger import CustomLogger


logger = CustomLogger()
"""The logger for the module."""

DataType = Literal["numerical", "categorical", "boolean", "datetime", "timedelta"]
"""Type alias for the data types."""


class DataSchema:
    """DataSchema class for storing type information about the dataset."""

    def __init__(
        self,
        numerical: list[str] | tuple[str, ...] | tuple[()] = (),
        categorical: list[str] | tuple[str, ...] | tuple[()] = (),
        boolean: list[str] | tuple[str, ...] | tuple[()] = (),
        datetime: list[str] | tuple[str, ...] | tuple[()] = (),
        timedelta: list[str] | tuple[str, ...] | tuple[()] = (),
    ) -> None:
        """Initialize DataSchema with the given features.

        Args:
            numerical: List of numerical features.
            categorical: List of categorical features.
            boolean: List of boolean features.
            datetime: List of datetime features.
            timedelta: List of timedelta features.

        Raises:
            ValueError: If feature names are not unique across all types.
        """
        if len(set(numerical) | set(categorical) | set(boolean) | set(datetime) | set(timedelta)) != sum(
            map(len, (numerical, categorical, boolean, datetime, timedelta))
        ):
            msg = "Feature names must be unique across all types."
            logger.error(msg)
            raise ValueError(msg)

        self.features: dict[DataType, list[str]] = {
            "numerical": sorted(list(numerical)),
            "categorical": sorted(list(categorical)),
            "boolean": sorted(list(boolean)),
            "datetime": sorted(list(datetime)),
            "timedelta": sorted(list(timedelta)),
        }

    def __eq__(self, other: object) -> bool:
        """Check if two DataSchema objects are equal.

        Args:
            other: Object to compare with.

        Returns:
            True if the two DataSchema objects are equal, False otherwise.
        """
        return isinstance(other, DataSchema) and DictFingerprinter().fingerprint(
            self.to_dict()
        ) == DictFingerprinter().fingerprint(other.to_dict())

    def __hash__(self) -> int:
        """Get the hash of the DataSchema.

        Warning:
            This method is not intended to be used for stable hashing across runs. Please
            refer to the `DictFingerprinter` class in the `mleko.utils.fingerprinter` module
            for stable hashing of the DataSchema.

        Returns:
            Hash of the DataSchema.
        """
        return hash(DictFingerprinter().fingerprint(self.to_dict()))

    def __repr__(self) -> str:
        """Get the string representation of DataSchema.

        Returns:
            String representation of DataSchema.
        """
        features_str = ", ".join(f"{dtype}={features}" for dtype, features in self.features.items() if features)
        return f"DataSchema({features_str})"

    def __str__(self) -> str:
        """Get the string representation of DataSchema.

        Returns:
            String representation of DataSchema.
        """
        return str(self.to_dict())

    def get_features(self, types: list[DataType] | tuple[DataType, ...] | tuple[()] = ()) -> list[str]:
        """Get features of a given type.

        If no type is specified, all features are returned.

        Args:
            types: List of data types to be returned.

        Returns:
            List of features of a given type.
        """
        if not types:
            return sum(self.features.values(), [])

        return sum((self.features[dtype] for dtype in types), [])

    def get_type(self, feature: str) -> DataType:
        """Get the type of a given feature.

        Args:
            feature: Feature name.

        Raises:
            ValueError: If feature is not found in the schema.

        Returns:
            Feature data type.
        """
        for dtype, features in self.features.items():
            if feature in features:
                return dtype

        msg = f"{feature} not found in the schema."
        logger.error(msg)
        raise ValueError(msg)

    def drop_features(self, features: set[str] | list[str] | tuple[str, ...] | tuple[()]) -> DataSchema:
        """Drop a feature from the DataSchema.

        Args:
            features: List of feature names to be dropped.
        """
        for dropped_feature in features:
            dtype = self.get_type(dropped_feature)
            self.features[dtype].remove(dropped_feature)

        return self

    def add_feature(self, feature: str, dtype: DataType) -> DataSchema:
        """Add a feature to the DataSchema.

        Args:
            feature: Feature name.
            dtype: Feature data type.

        Raises:
            ValueError: If feature is already present in the schema.
        """
        if feature in self.get_features():
            msg = f"{feature} already present in the schema."
            logger.error(msg)
            raise ValueError(msg)

        self.features[dtype].append(feature)
        self.features[dtype].sort()

        return self

    def change_feature_type(self, feature: str, dtype: DataType) -> DataSchema:
        """Change the type of a feature in the DataSchema.

        Args:
            feature: Feature name.
            dtype: Feature data type.

        Raises:
            ValueError: If feature is not present in the schema.
        """
        if feature not in self.get_features():
            msg = f"{feature} not present in the schema."
            logger.error(msg)
            raise ValueError(msg)

        self.drop_features([feature])
        self.add_feature(feature, dtype)

        return self

    def to_dict(self) -> dict[str, list[str]]:
        """Return the dict representation of DataSchema.

        Returns:
            Dict representation of DataSchema.
        """
        return {dtype: features for dtype, features in self.features.items()}

    def copy(self) -> DataSchema:
        """Create a copy of this DataSchema.

        Returns:
            A copy of this DataSchema.
        """
        copied_features = {dtype: copy.deepcopy(features) for dtype, features in self.features.items()}
        data_schema_copy = DataSchema(**copied_features)

        return data_schema_copy



"""Module containing the FeatureSelectStep class.""" ""
from __future__ import annotations

from typing import Union, cast

from typing_extensions import TypedDict
from vaex import DataFrame

from mleko.dataset.data_schema import DataSchema
from mleko.dataset.feature_select.base_feature_selector import BaseFeatureSelector
from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import FitTransformAction, FitTransformPipelineStep
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr


logger = CustomLogger()
"""The logger for the module."""


class FeatureSelectStepInputType(TypedDict):
    """The input type of the FeatureSelectStep."""

    data_schema: Union[str, DataSchema]
    """DataSchema or the key for the DataSchema to be used for feature selection."""

    dataframe: Union[str, DataFrame]
    """DataFrame or the key for the DataFrame to be used for feature selection."""


class FeatureSelectStepOutputFitType(TypedDict):
    """The output type of the FeatureSelectStep when action is 'fit'."""

    data_schema: str
    """The key for the updated DataSchema after feature selection."""

    feature_selector: str
    """The key for the fitted feature selector after feature selection."""


class FeatureSelectStepOutputTransformType(TypedDict):
    """The output type of the FeatureSelectStep when action is 'transform'."""

    data_schema: str
    """The key for the updated DataSchema after feature selection."""

    dataframe: str
    """The key for the transformed DataFrame after feature selection."""


class FeatureSelectStepOutputFitTransformType(FeatureSelectStepOutputFitType, FeatureSelectStepOutputTransformType):
    """The output type of the FeatureSelectStep when action is 'fit_transform'."""

    pass


class FeatureSelectStep(FitTransformPipelineStep):
    """Pipeline step that selects a subset of features from a DataFrame."""

    _inputs: FeatureSelectStepInputType
    _outputs: (
        FeatureSelectStepOutputFitType | FeatureSelectStepOutputTransformType | FeatureSelectStepOutputFitTransformType
    )

    @auto_repr
    def __init__(
        self,
        feature_selector: BaseFeatureSelector,
        action: FitTransformAction,
        inputs: FeatureSelectStepInputType,
        outputs: (
            FeatureSelectStepOutputFitType
            | FeatureSelectStepOutputTransformType
            | FeatureSelectStepOutputFitTransformType
        ),
        cache_group: str | None = None,
    ) -> None:
        """Initialize the FeatureSelectStep with the specified feature selector.

        Args:
            feature_selector: The FeatureSelector responsible for handling feature selection.
            action: The action to perform, one of "fit", "transform", or "fit_transform".
            inputs: A dictionary of input keys following the `FeatureSelectStepInputType` schema.
            outputs: A dictionary of output keys following the `FeatureSelectStepOutputFitType`,
                `FeatureSelectStepOutputTransformType`, or `FeatureSelectStepOutputFitTransformType` schema depending
                on the action.
            cache_group: The cache group to use.
        """
        super().__init__(action, inputs, outputs, cache_group)
        self._feature_selector = feature_selector

    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Perform feature selection using the configured feature selector.

        Args:
            data_container: Contains the input DataFrame.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Returns:
            A DataContainer containing the result depending on the action.
        """
        data_schema = self._validate_and_get_input(self._inputs["data_schema"], DataSchema, data_container)
        dataframe = self._validate_and_get_input(self._inputs["dataframe"], DataFrame, data_container)

        if self._action == "fit":
            self._outputs = cast(FeatureSelectStepOutputFitType, self._outputs)
            ds, feature_selector = self._feature_selector.fit(
                data_schema, dataframe, self._cache_group, force_recompute
            )
            data_container.data[self._outputs["data_schema"]] = ds
            data_container.data[self._outputs["feature_selector"]] = feature_selector
        elif self._action == "transform":
            self._outputs = cast(FeatureSelectStepOutputTransformType, self._outputs)
            ds, df = self._feature_selector.transform(data_schema, dataframe, self._cache_group, force_recompute)
            data_container.data[self._outputs["data_schema"]] = ds
            data_container.data[self._outputs["dataframe"]] = df
        elif self._action == "fit_transform":
            self._outputs = cast(FeatureSelectStepOutputFitTransformType, self._outputs)
            ds, feature_selector, df = self._feature_selector.fit_transform(
                data_schema, dataframe, self._cache_group, force_recompute
            )
            data_container.data[self._outputs["data_schema"]] = ds
            data_container.data[self._outputs["feature_selector"]] = feature_selector
            data_container.data[self._outputs["dataframe"]] = df
        return data_container

    def _get_input_model(self) -> type[FeatureSelectStepInputType]:
        """Get the input type for the FeatureSelectStep.

        Returns:
            Input type for the FeatureSelectStep.
        """
        return FeatureSelectStepInputType

    def _get_output_model(
        self,
    ) -> type[
        FeatureSelectStepOutputFitType | FeatureSelectStepOutputTransformType | FeatureSelectStepOutputFitTransformType
    ]:
        """Get the output type for the FeatureSelectStep.

        Returns:
            Output type for the FeatureSelectStep.
        """
        if self._action == "fit":
            return FeatureSelectStepOutputFitType
        if self._action == "transform":
            return FeatureSelectStepOutputTransformType
        return FeatureSelectStepOutputFitTransformType



"""A module for filtering `vaex` DataFrames using `imblearn` sampling methods."""

from __future__ import annotations

import logging
from pathlib import Path

import numpy as np
import pyarrow as pa
import vaex
from imblearn.under_sampling.base import BaseSampler

from mleko.cache.fingerprinters.dict_fingerprinter import DictFingerprinter
from mleko.cache.fingerprinters.vaex_fingerprinter import VaexFingerprinter
from mleko.cache.handlers.vaex_cache_handler import VAEX_DATAFRAME_CACHE_HANDLER
from mleko.dataset.data_schema import DataSchema
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr
from mleko.utils.vaex_helpers import get_column, get_columns, get_indices

from .base_filter import BaseFilter


logger = CustomLogger()
"""A module-level logger instance."""


class ImblearnResamplingFilter(BaseFilter):
    """A class that handles filtering of `vaex` DataFrames using `imblearn` sampling methods."""

    @auto_repr
    def __init__(
        self,
        sampler: BaseSampler,
        target_column: str,
        random_state: int | None = 42,
        verbosity: int = logging.INFO,
        cache_directory: str | Path = "data/imblearn-sampling-filter",
        cache_size: int = 1,
    ) -> None:
        """Initializes the `ImblearnResamplingFilter` with the given `imblearn` sampler and target column.

        The `imblearn` sampler should be a sample object that inherits from `BaseSampler`.
        For example, `imblearn.under_sampling.RandomUnderSampler`. Refer to the `imblearn` documentation for more
        information (https://imbalanced-learn.org/stable/introduction.html).

        Args:
            sampler: The `imblearn` sampler to be used for sampling.
            target_column: The target column to be used for sampling.
            random_state: The random state to be used for reproducibility, and will recursively set the random state
                of the sampler and all nested objects if set.
            verbosity: The verbosity level of the logger.
            cache_directory: The target directory where the filtered dataframes are to be saved.
            cache_size: The maximum number of cache entries.

        Examples:
            >>> import vaex
            >>> from imblearn.under_sampling import RandomUnderSampler
            >>> from mleko.data.filter import ImblearnUnderSamplingFilter
            >>> df = vaex.from_arrays(x=[1, 2, 3], y=[4, 5, 6])
            >>> sampler = RandomUnderSampler()
            >>> filter = ImblearnUnderSamplingFilter(sampler=sampler, target_column="x")
            >>> df_filtered = filter.filter(df)
            >>> df_filtered
                #    x    y
                0    1    4
                1    2    5
        """
        super().__init__(cache_directory, cache_size)
        self._sampler = sampler
        self._target_column = target_column
        self._random_state = random_state
        self._reset_random_state()
        logger.set_level(verbosity)

    def filter(
        self,
        data_schema: DataSchema,
        dataframe: vaex.DataFrame,
        cache_group: str | None = None,
        force_recompute: bool = False,
        disable_cache: bool = False,
    ) -> vaex.DataFrame:
        """Filters the given DataFrame using the specified `imblearn` sampler.

        Args:
            data_schema: Data schema of the DataFrame.
            dataframe: DataFrame to be filtered.
            cache_group: The cache group to use.
            force_recompute: Forces recomputation if True, otherwise reads from the cache if available.
            disable_cache: If set to True, disables the cache.

        Returns:
            The filtered DataFrame.
        """
        return self._cached_execute(
            lambda_func=lambda: self._filter(data_schema, dataframe),
            cache_key_inputs=[
                self._sampler.__class__.__qualname__,
                (self._sampler.get_params(deep=True), DictFingerprinter()),
                self._target_column,
                (data_schema.to_dict(), DictFingerprinter()),
                (dataframe, VaexFingerprinter()),
            ],
            cache_group=cache_group,
            force_recompute=force_recompute,
            cache_handlers=VAEX_DATAFRAME_CACHE_HANDLER,
            disable_cache=disable_cache,
        )

    def _filter(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> vaex.DataFrame:
        """Filters the given DataFrame using the specified `imblearn` sampler.

        Args:
            data_schema: Data schema of the DataFrame.
            dataframe: DataFrame to be filtered.

        Returns:
            The filtered DataFrame.
        """
        resampler_name = self._sampler.__class__.__qualname__
        logger.info(f"Imblearn resampling DataFrame using {resampler_name!r}.")
        df: vaex.DataFrame = dataframe.copy()

        x = get_columns(df, data_schema.get_features()).to_pandas_df()
        y = get_column(df, self._target_column).to_numpy()
        resampled_x, resampled_y = self._sampler.fit_resample(x, y)  # type: ignore
        resampled_x_len = resampled_x.shape[0]
        original_df_len = df.shape[0]

        # Over-Sampling
        if resampled_x_len > original_df_len:
            resampled_df = vaex.from_pandas(resampled_x.iloc[original_df_len:])
            resampled_df_len = resampled_df.shape[0]

            logger.info("Ensuring that the resampled DataFrame handles NaN values correctly.")
            for col_name in data_schema.get_features():
                column = get_column(resampled_df, col_name)
                if column.countnan() > 0:
                    column_values = column.values
                    resampled_df[col_name] = pa.array(column_values, mask=np.isnan(column_values))  # type: ignore
            resampled_df[self._target_column] = resampled_y[original_df_len:]

            synthetic_column_name = "Synthetic"
            logger.info(f"Setting the {synthetic_column_name!r} column to 1 for the synthetic samples.")
            resampled_df[synthetic_column_name] = np.ones(resampled_df_len, dtype="int8")
            df[synthetic_column_name] = np.zeros(original_df_len, dtype="int8")
            resampled_df = vaex.concat([df, resampled_df])
            logger.info(f"Generated {resampled_df_len} synthetic samples.")

        # Under-Sampling
        elif resampled_x_len < original_df_len:
            resampled_df = get_indices(df, resampled_x.index)  # type: ignore
            logger.info(f"Removed {original_df_len - resampled_x_len} samples using {resampler_name!r}.")

        # No Resampling
        else:  # pragma: no cover
            resampled_df = get_indices(df, resampled_x.index)  # type: ignore
            logger.info("No resampling was performed as the DataFrame size remains the same.")

        self._reset_random_state()
        return resampled_df

    def _reset_random_state(self) -> None:
        """Resets the random state of the sampler."""
        self._sampler.set_params(
            **{
                param: self._random_state
                for param in self._sampler.get_params().keys()
                if param.endswith("random_state")
            }
        )



"""This module contains the `ExpressionFilter` class, which is used to filter `vaex` DataFrames."""

from __future__ import annotations

from pathlib import Path

import vaex

from mleko.cache.fingerprinters import DictFingerprinter, VaexFingerprinter
from mleko.cache.handlers.vaex_cache_handler import VAEX_DATAFRAME_CACHE_HANDLER
from mleko.dataset.data_schema import DataSchema
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr

from .base_filter import BaseFilter


logger = CustomLogger()
"""A module-level logger instance."""


class ExpressionFilter(BaseFilter):
    """A class that handles filtering of `vaex` DataFrames based on a given `vaex` expression."""

    @auto_repr
    def __init__(
        self,
        expression: str,
        cache_directory: str | Path = "data/expression-filter",
        cache_size: int = 1,
    ) -> None:
        """Initializes the `ExpressionFilter` with the given expression.

        The expression should be a valid Vaex expression that evaluates to a boolean value.

        Note:
            To filter by a date column, use the `scalar_datetime` function. For example, to filter by a date column
            named `date` and return the rows before `2020-06-01`, use the
            expression `"date < scalar_datetime('2020-06-01')"`.

        Args:
            expression: The expression to be used for filtering.
            cache_directory: The target directory where the filtered dataframes are to be saved.
            cache_size: The maximum number of cache entries.

        Example:
            >>> import vaex
            >>> from mleko.data.filter import ExpressionFilter
            >>> df = vaex.from_arrays(x=[1, 2, 3], y=[4, 5, 6])
            >>> filter = ExpressionFilter(expression="x > 1")
            >>> df_filtered = filter.filter(df)
            >>> df_filtered
                #    x    y
                0    2    5
                1    3    6
        """
        super().__init__(cache_directory, cache_size)
        self._expression = expression

    def filter(
        self,
        data_schema: DataSchema,
        dataframe: vaex.DataFrame,
        cache_group: str | None = None,
        force_recompute: bool = False,
        disable_cache: bool = False,
    ) -> vaex.DataFrame:
        """Filter the given dataframe based on the expression.

        Args:
            data_schema: The data schema to be used for filtering.
            dataframe: The dataframe to be filtered.
            cache_group: The cache group to use.
            force_recompute: Forces recomputation if True, otherwise reads from the cache if available.
            disable_cache: If set to True, disables the cache.

        Returns:
            The filtered dataframe.
        """
        return self._cached_execute(
            lambda_func=lambda: self._filter(data_schema, dataframe),
            cache_key_inputs=[
                self._expression,
                (data_schema.to_dict(), DictFingerprinter()),
                (dataframe, VaexFingerprinter()),
            ],
            cache_group=cache_group,
            force_recompute=force_recompute,
            cache_handlers=VAEX_DATAFRAME_CACHE_HANDLER,
            disable_cache=disable_cache,
        )

    def _filter(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> vaex.DataFrame:
        """Filter the given dataframe based on the expression.

        Args:
            dataframe: The dataframe to be filtered.

        Returns:
            The filtered dataframe.
        """
        logger.info(f"Filtering dataframe based on expression {self._expression!r}.")
        filtered_df = dataframe.filter(f"({self._expression})").extract()
        logger.info(
            f"Filtered dataframe into shape {filtered_df.shape}, "
            f"dropped {dataframe.shape[0] - filtered_df.shape[0]} rows."
        )
        return filtered_df



"""Module for the frequency encoder transformer."""

from __future__ import annotations

from pathlib import Path
from typing import Hashable, Literal

import vaex
import vaex.ml

from mleko.dataset.data_schema import DataSchema
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr

from .base_transformer import BaseTransformer


logger = CustomLogger()
"""A module-level logger for the module."""


class FrequencyEncoderTransformer(BaseTransformer):
    """Transforms features using frequency encoding."""

    @auto_repr
    def __init__(
        self,
        features: list[str] | tuple[str, ...],
        unseen_strategy: Literal["zero", "nan"] = "nan",
        cache_directory: str | Path = "data/frequency-encoder-transformer",
        cache_size: int = 1,
    ) -> None:
        """Initializes the transformer.

        Uses the `vaex.ml.FrequencyEncoder` transformer, which encodes categorical features using the frequency of
        their respective samples. If a value is not seen during fitting, it will be encoded as zero or nan,
        depending on the `unseen_strategy` parameter. Missing values will be encoded as nan, but will still count
        towards the frequency of other values.

        Warning:
            Should only be used with categorical features. High cardinality features are not recommended as they will
            result in very small frequencies.

        Args:
            features: List of feature names to be used by the transformer.
            unseen_strategy: Strategy to use for unseen values once the transformer is fitted.
            cache_directory: Directory where the cache will be stored locally.
            cache_size: The maximum number of entries to keep in the cache.

        Examples:
            >>> import vaex
            >>> from mleko.dataset.transform import FrequencyEncoderTransformer
            >>> from mleko.utils.vaex_helpers import get_column
            >>> df = vaex.from_arrays(
            ...     a=["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"],
            ...     b=["a", "a", "a", "a", None, None, None, None, None, None],
            ...     c=["a", "b", "b", "b", "b", "b", None, None, None, None],
            ... )
            >>> ds = DataSchema(
            ...     categorical=["a", "b", "c"],
            ... )
            >>> _, _, df = FrequencyEncoderTransformer(
            ...     features=["a", "b"],
            ... ).fit_transform(ds, df)
            >>> df["a"].tolist()
            [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
            >>> df["b"].tolist()
            [0.4, 0.4, 0.4, 0.4, nan, nan, nan, nan, nan, nan]
        """
        super().__init__(features, cache_directory, cache_size)
        self._unseen_strategy = unseen_strategy
        self._transformer = vaex.ml.FrequencyEncoder(
            features=self._features, unseen_strategy=self._unseen_strategy, prefix=""
        )

    def _fit(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, vaex.ml.FrequencyEncoder]:
        """Fits the transformer on the input data.

        Args:
            data_schema: The DataSchema of the DataFrame.
            dataframe: The DataFrame to fit the transformer on.

        Returns:
            Updated DataSchema and the fitted transformer.
        """
        logger.info(f"Fitting frequency encoder transformer ({len(self._features)}): {self._features}.")
        self._transformer.fit(dataframe)

        ds = data_schema.copy()
        for feature in self._features:
            ds = ds.change_feature_type(feature, "numerical")

        return ds, self._transformer

    def _transform(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, vaex.DataFrame]:
        """Transforms the features in the DataFrame using frequency encoding.

        Args:
            data_schema: The DataSchema of the DataFrame.
            dataframe: The DataFrame to transform.

        Returns:
            Updated DataSchema and the transformed DataFrame.
        """
        logger.info(f"Transforming features using frequency encoding ({len(self._features)}): {self._features}.")
        transformed_df = self._transformer.transform(dataframe)

        ds = data_schema.copy()
        for feature in self._features:
            ds = ds.change_feature_type(feature, "numerical")

        return ds, transformed_df

    def _fingerprint(self) -> Hashable:
        """Returns the fingerprint of the transformer.

        Append the `unseen_strategy` to the fingerprint.

        Returns:
            A hashable object that uniquely identifies the transformer.
        """
        return super()._fingerprint(), self._unseen_strategy



"""Module containing the TuneStep class."""

from __future__ import annotations

from typing import Union

from typing_extensions import TypedDict
from vaex import DataFrame

from mleko.dataset.data_schema import DataSchema
from mleko.model.tune.base_tuner import BaseTuner
from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import PipelineStep
from mleko.utils.decorators import auto_repr


class TuneStepInputType(TypedDict):
    """The input type of the TuneStep."""

    data_schema: Union[str, DataSchema]
    """DataSchema or the key for the DataSchema to be used for hyperparameter tuning."""

    dataframe: Union[str, DataFrame]
    """DataFrame or the key for the DataFrame to be used for hyperparameter tuning."""


class TuneStepOutputType(TypedDict):
    """The output type of the TuneStep."""

    hyperparameters: str
    """The key for the best hyperparameters after tuning."""

    score: str
    """The key for the best objective score after tuning."""

    metadata: str
    """The key for the optional metadata object after tuning."""


class TuneStep(PipelineStep):
    """Pipeline step for hyperparameter tuning."""

    _inputs: TuneStepInputType
    _outputs: TuneStepOutputType

    @auto_repr
    def __init__(
        self,
        tuner: BaseTuner,
        inputs: TuneStepInputType,
        outputs: TuneStepOutputType,
        cache_group: str | None = None,
    ) -> None:
        """Initialize the TuneStep with the specified tuner.

        Args:
            tuner: The tuner used for hyperparameter tuning.
            inputs: A dictionary of input keys following the `TuneStepInputType` schema.
            outputs: A dictionary of output keys following the `TuneStepOutputType` schema.
            cache_group: The cache group to use.
        """
        super().__init__(inputs, outputs, cache_group)
        self._tuner = tuner

    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Perform hyperparameter tuning.

        Args:
            data_container: Contains the input.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Returns:
            A DataContainer containing the output of the tuning step. It contains the best hyperparameters, the best
            objective score, and an optional metadata object.
        """
        data_schema = self._validate_and_get_input(self._inputs["data_schema"], DataSchema, data_container)
        dataframe = self._validate_and_get_input(self._inputs["dataframe"], DataFrame, data_container)

        hyperparameters, score, metadata = self._tuner.tune(data_schema, dataframe, self._cache_group, force_recompute)
        data_container.data[self._outputs["hyperparameters"]] = hyperparameters
        data_container.data[self._outputs["score"]] = score
        data_container.data[self._outputs["metadata"]] = metadata
        return data_container

    def _get_input_model(self) -> type[TuneStepInputType]:
        """Get the input model for the TuneStep.

        Returns:
            The input model for the TuneStep.
        """
        return TuneStepInputType

    def _get_output_model(self) -> type[TuneStepOutputType]:
        """Get the output model for the TuneStep.

        Returns:
            The output model for the TuneStep.
        """
        return TuneStepOutputType



"""This module contains the basic `CacheMixin` class for caching the results of method calls.

This class can be used as a mixin to add caching functionality to a class. It provides the basic
functionality for caching the results of method calls based on user-defined cache keys and fingerprints.

Combining this class with the format mixins can be used to add support for caching different data
formats, such as Vaex DataFrames in Arrow format.
"""

from __future__ import annotations

import hashlib
import inspect
import pickle
import re
from pathlib import Path
from typing import Any, Callable, Hashable, Sequence

from mleko.cache.fingerprinters.base_fingerprinter import BaseFingerprinter
from mleko.cache.handlers import PICKLE_CACHE_HANDLER, CacheHandler
from mleko.utils.custom_logger import CustomLogger


logger = CustomLogger()
"""A module-level logger instance."""


def get_qualified_name_from_frame(frame: inspect.FrameInfo) -> str:
    """Gets the fully qualified name of the function or method associated with the provided frame.

    Args:
        frame: A `FrameInfo` object containing the information of the function or method call.

    Returns:
        A string representing the fully qualified name, in the format "module.class.method" for class methods or
        "module.function" for functions.
    """
    caller_function = frame.function
    caller_obj = inspect.getmodule(frame[0])
    module_name = caller_obj.__name__ if caller_obj is not None else "__main__"
    if "self" in frame.frame.f_locals:
        class_name = frame.frame.f_locals["self"].__class__.__name__
        return f"{module_name}.{class_name}.{caller_function}"
    return f"{module_name}.{caller_function}"


def get_qualified_name_of_caller(frame_depth: int) -> str:
    """Gets the fully qualified name of the calling function or method.

    The fully qualified name is in the format "module.class.method" for class methods or "module.function" for
    functions.

    Args:
        frame_depth: The depth of the frame to inspect. The default value is 2, which is the frame of the calling
            function or method. For each nested function or method, the frame depth should be increased by 1.

    Returns:
        A string representing the fully qualified name of the calling function or method.
    """
    frame_qualname = get_qualified_name_from_frame(inspect.stack()[frame_depth])
    class_method_name = ".".join(frame_qualname.split(".")[-2:])
    return class_method_name


class CacheMixin:
    """A mixin class for caching the results of method calls based on user-defined cache keys and fingerprints.

    The basic functionality of this class is to cache the results of method calls based on user-defined cache keys and
    fingerprints. The cache keys can be a mix of hashable values and tuples containing a value and a BaseFingerprinter
    instance for generating fingerprints. The `CacheMixin` class will save cache files in the specified cache directory
    using the cache key as the filename and the cache file suffix as the file extension. The cache files will be saved
    in the cache directory as pickle files.

    Warning:
        This class maintains an ever-growing cache, which means that the cache size may increase indefinitely
        with new method calls, possibly consuming a large amount of disk space. It does not implement any
        cache eviction strategy. It is recommended to either clear the cache manually when needed or
        use the LRUCacheMixin class, which extends this class to provide an LRU cache mechanism with
        eviction of least recently used cache entries based on a specified maximum number of cache entries.
    """

    def __init__(self, cache_directory: str | Path, disable_cache: bool) -> None:
        """Initializes the `CacheMixin` with the provided cache directory.

        Note:
            The cache directory will be created if it does not exist.

        Args:
            cache_directory: The directory where cache files will be stored.
            disable_cache: Whether to disable the cache.

        Examples:
            >>> from mleko.cache.cache_mixin import CacheMixin
            >>> class MyClass(CacheMixin):
            ...     def __init__(self):
            ...         super().__init__(".cache", "pkl")
            ...
            ...     def my_method(self, x):
            ...         return self._cached_execute(lambda: x ** 2, [x])
            ...
            >>> my_class = MyClass()
            >>> my_class.my_method(2)
            4 # This will be computed and cached
            >>> my_class.my_method(2)
            4 # This will be loaded from the cache
            >>> my_class.my_method(3)
            9 # This will be recomputed and cached
        """
        self._cache_directory = Path(cache_directory)
        self._cache_directory.mkdir(parents=True, exist_ok=True)
        self._cache_type_name = self._find_cache_type_name(self.__class__)
        self._disable_cache = disable_cache

    def _cached_execute(
        self,
        lambda_func: Callable[[], Any],
        cache_key_inputs: list[Hashable | tuple[Any, BaseFingerprinter]],
        cache_group: str | None = None,
        force_recompute: bool = False,
        cache_handlers: CacheHandler | list[CacheHandler] | None = None,
        disable_cache: bool = False,
    ) -> Any:
        """Executes the given function, caching the results based on the provided cache keys and fingerprints.

        Warning:
            The cache group is used to group related cache keys together to prevent collisions between cache keys
            originating from the same method. For example, if a method is called during the training and testing
            phases of a machine learning pipeline, the cache keys for the training and testing phases should be
            using different cache groups to prevent collisions between the cache keys for the two phases. Otherwise,
            the later cache keys might overwrite the earlier cache entries.

        Args:
            lambda_func: A lambda function to execute.
            cache_key_inputs: A list of cache keys that can be a mix of hashable values and tuples containing
                a value and a BaseFingerprinter instance for generating fingerprints.
            cache_group: A string representing the cache group, used to group related cache keys together when methods
                are called independently.
            force_recompute: A boolean indicating whether to force recompute the result and update the cache, even if a
                cached result is available.
            cache_handlers: A CacheHandler instance or a list of CacheHandler instances. If None, the cache files will
                be read using pickle. If a single CacheHandler instance is provided, it will be used for all cache
                files. If a list of CacheHandler instances is provided, each CacheHandler instance will be used for
                each cache file.
            disable_cache: Overrides the class-level `disable_cache` attribute. If set to True, disables the cache.

        Returns:
            A tuple containing a boolean indicating whether the cached result was used, and the result of executing the
            given function. If a cached result is available and `force_recompute` is False, the cached result will be
            returned instead of recomputing the result.
        """
        if self._disable_cache or disable_cache:
            return lambda_func()

        if cache_handlers is None:
            cache_handlers = PICKLE_CACHE_HANDLER

        class_method_name = get_qualified_name_of_caller(2)
        cache_key = self._compute_cache_key(cache_key_inputs, cache_group)

        if not force_recompute:
            output = self._load_from_cache(cache_key, cache_handlers)
            if output is not None:
                logger.info(
                    f"\033[32mCache Hit\033[0m ({self._cache_type_name}) {class_method_name}: Using cached output."
                )
                return output
            else:
                logger.info(
                    f"\033[31mCache Miss\033[0m ({self._cache_type_name}) {class_method_name}: Executing method."
                )
        else:
            logger.info(
                f"\033[33mForce Cache Refresh\033[0m ({self._cache_type_name}) {class_method_name}: Executing method."
            )

        output = lambda_func()
        self._save_to_cache(cache_key, output, cache_handlers)
        return self._load_from_cache(cache_key, cache_handlers)

    def _compute_cache_key(
        self,
        cache_key_inputs: list[Hashable | tuple[Any, BaseFingerprinter]],
        cache_group: str | None = None,
        frame_depth: int = 3,
    ) -> str:
        """Computes the cache key based on the provided cache keys and the calling function's fully qualified name.

        Args:
            cache_key_inputs: A list of cache keys that can be a mix of hashable values and tuples containing a
                value and a BaseFingerprinter instance for generating fingerprints.
            cache_group: A string representing the cache group.
            frame_depth: The depth of the frame to inspect. The default value is 2, which is the frame of the calling
                function or method. For each nested function or method, the frame depth should be increased by 1.

        Raises:
            ValueError: If the computed cache key is too long.

        Returns:
            A string representing the computed cache key, which is the MD5 hash of the fully qualified name of the
            calling function or method, along with the fingerprints of the provided cache keys.
        """
        values_to_hash: list[Hashable] = []

        for key_input in cache_key_inputs:
            if isinstance(key_input, tuple) and len(key_input) == 2 and isinstance(key_input[1], BaseFingerprinter):
                value, fingerprinter = key_input
                values_to_hash.append(fingerprinter.fingerprint(value))
            else:
                values_to_hash.append(key_input)

        logger.debug(f"Cache key inputs: {values_to_hash}")
        data = pickle.dumps(values_to_hash)
        cache_key_prefix = get_qualified_name_of_caller(frame_depth)
        if cache_group is not None:
            cache_key_prefix = f"{cache_key_prefix}.{cache_group}"

        cache_key = f"{cache_key_prefix}.{hashlib.md5(data).hexdigest()}"
        if len(cache_key) > 235:
            msg = (
                f"The computed cache key is too long ({len(cache_key)} chars)."
                "The maximum length of a cache key is 235 chars, and given the current class, the maximum "
                f"length of the provided cache_group is {235 - len(cache_key)} chars. "
                "Please reduce the length of the cache_group."
            )
            logger.error(msg)
            raise ValueError(msg)

        return f"{cache_key_prefix}.{hashlib.md5(data).hexdigest()}"

    def _load_from_cache(
        self,
        cache_key: str,
        cache_handlers: CacheHandler | list[CacheHandler],
    ) -> Any | None:
        """Loads data from the cache based on the provided cache key.

        Args:
            cache_key: A string representing the cache key.
            cache_handlers: A CacheHandler instance or a list of CacheHandler instances. If a single CacheHandler
                instance is provided, it will be used for all cache files. If a list of CacheHandler instances is
                provided, each CacheHandler instance will be used for each cache file.

        Returns:
            The cached data if it exists, or None if there is no data for the given cache key.
        """

        def extract_number(file_path: Path) -> int:
            result = re.search(r"[a-fA-F\d]{32}_(\d+).", str(file_path))
            return int(result.group(1)) if result else 0

        cache_handler_suffixes = (
            {f".{cache_handler.suffix}" for cache_handler in cache_handlers}
            if isinstance(cache_handlers, list)
            else {f".{cache_handlers.suffix}"}
        ).union({f".{PICKLE_CACHE_HANDLER.suffix}"})

        cache_file_paths = [
            f
            for f in sorted(list(self._cache_directory.glob(f"{cache_key}*.*")), key=extract_number)
            if f.suffix in cache_handler_suffixes
        ]

        if cache_file_paths:
            output_data = []
            for i, cache_file_path in enumerate(cache_file_paths):
                handler = self._get_handler(cache_handlers, i)
                if cache_file_path.suffix != f".{handler.suffix}":
                    handler = PICKLE_CACHE_HANDLER

                output_data.append(handler.reader(cache_file_path))
            return tuple(output_data) if len(output_data) > 1 else output_data[0]
        return None

    def _get_handler(self, cache_handlers: CacheHandler | list[CacheHandler], index: int = 0) -> CacheHandler:
        """Gets the cache handler at the given index.

        Args:
            cache_handlers: A CacheHandler instance or a list of CacheHandler instances.
            index: The index of the cache handler to get.

        Returns:
            Handler at the given index. If a single CacheHandler instance is provided, it will be returned.
        """
        if isinstance(cache_handlers, list):
            return cache_handlers[index]
        return cache_handlers

    def _save_to_cache(
        self,
        cache_key: str,
        output: Any | Sequence[Any],
        cache_handlers: CacheHandler | list[CacheHandler],
    ) -> None:
        """Saves the given data to the cache using the provided cache key.

        If the output is a sequence, each element will be saved to a separate cache file. Otherwise, the output will be
        saved to a single cache file. The cache file will be saved in the cache directory with the cache key as the
        filename and the cache file suffix as the file extension.

        Args:
            cache_key: A string representing the cache key.
            output: The data to be saved to the cache.
            cache_handlers: A CacheHandler instance or a list of CacheHandler instances. If a single CacheHandler
                instance is provided, it will be used for all cache files. If a list of CacheHandler instances is
                provided, each CacheHandler instance will be used for each cache file.
        """
        if isinstance(output, Sequence):
            for i, output_item in enumerate(output):
                self._write_to_cache_file(cache_key, output_item, i, cache_handlers, is_sequence_output=True)
        else:
            self._write_to_cache_file(cache_key, output, 0, cache_handlers, is_sequence_output=False)

    def _write_to_cache_file(
        self,
        cache_key: str,
        output_item: Any,
        index: int,
        cache_handlers: CacheHandler | list[CacheHandler],
        is_sequence_output: bool,
    ) -> None:
        """Writes the given data to the cache file using the provided cache key.

        If the output is None and the cache handler cannot handle None, the output will be saved using the pickle
        cache handler. Otherwise, the output will be saved to a cache file using the provided cache handler.

        Args:
            cache_key: A string representing the cache key.
            output_item: The data to be saved to the cache.
            index: The index of the cache handler to use.
            cache_handlers: A CacheHandler instance or a list of CacheHandler instances.
            is_sequence_output: Whether the output is a sequence or not. If True, the cache file will be saved with the
                index appended to the cache key.
        """
        handler = self._get_handler(cache_handlers, index)

        file_suffix = f"_{index}" if is_sequence_output else ""
        if output_item is None and not handler.can_handle_none:
            cache_file_path = self._cache_directory / f"{cache_key}{file_suffix}.{PICKLE_CACHE_HANDLER.suffix}"
            PICKLE_CACHE_HANDLER.writer(cache_file_path, output_item)
        else:
            cache_file_path = self._cache_directory / f"{cache_key}{file_suffix}.{handler.suffix}"
            handler.writer(cache_file_path, output_item)

    def _find_cache_type_name(self, cls: type) -> str | None:
        """Recursively searches the class hierarchy for the name of the class that inherits from `CacheMixin`.

        Args:
            cls: The class to search.

        Returns:
            The name of the class that inherits from `CacheMixin`, or None if no such class exists.
        """
        if CacheMixin.__name__ in cls.__name__:
            return cls.__name__.replace("Mixin", "")

        for base in cls.__bases__:
            found_class_name = self._find_cache_type_name(base)
            if found_class_name:
                return found_class_name
        return None  # pragma: no cover



"""The module provides `ExpressionSplitter` class for splitting Vaex DataFrames based on a given expression."""

from __future__ import annotations

from pathlib import Path

import vaex

from mleko.cache.fingerprinters import VaexFingerprinter
from mleko.cache.handlers.vaex_cache_handler import VAEX_DATAFRAME_CACHE_HANDLER
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr

from .base_splitter import BaseSplitter


logger = CustomLogger()
"""A module-level logger instance."""


class ExpressionSplitter(BaseSplitter):
    """A class that handles splitting of `vaex` DataFrames based on a given `vaex` expression."""

    @auto_repr
    def __init__(
        self,
        expression: str,
        cache_directory: str | Path = "data/expression-splitter",
        cache_size: int = 1,
    ) -> None:
        """Initializes the `ExpressionSplitter` with the given parameters.

        The expression should be a valid Vaex expression that evaluates to a boolean
        value. The rows for which the expression evaluates to True will be returned as the first dataframe,
        and the remaining rows will be returned as the second dataframe.

        Note:
            To filter by a date column, use the `scalar_datetime` function. For example, to filter by a date column
            named `date` and return the rows before `2020-06-01`, use the
            expression `"date < scalar_datetime('2020-06-01')"`.

        Args:
            expression: A valid Vaex expression that evaluates to a boolean value. The rows for which the expression
                evaluates to True will be returned as the first dataframe, and the remaining rows will be returned
                as the second dataframe.
            cache_directory: The target directory where the split dataframes are to be saved.
            cache_size: The maximum number of entries to keep in the cache.

        Example:
            >>> import vaex
            >>> from mleko.data.split import ExpressionSplitter
            >>> df = vaex.from_arrays(x=[1, 2, 3], y=[4, 5, 6])
            >>> splitter = ExpressionSplitter(expression="x > 1")
            >>> df_train, df_test = splitter.split(df)
            >>> df_train
                #    x    y
                0    2    5
                1    3    6
            >>> df_test
                #    x    y
                0    1    4
        """
        super().__init__(cache_directory, cache_size)
        self._expression = expression

    def split(
        self,
        dataframe: vaex.DataFrame,
        cache_group: str | None = None,
        force_recompute: bool = False,
        disable_cache: bool = False,
    ) -> tuple[vaex.DataFrame, vaex.DataFrame]:
        """Split the given dataframe into two parts.

        Args:
            dataframe: The dataframe to be split.
            cache_group: The cache group to use.
            force_recompute: Forces recomputation if True, otherwise reads from the cache if available.
            disable_cache: If set to True, disables the cache.

        Returns:
            A tuple containing the split dataframes.
        """
        return self._cached_execute(
            lambda_func=lambda: self._split(dataframe),
            cache_key_inputs=[
                self._expression,
                (dataframe, VaexFingerprinter()),
            ],
            cache_group=cache_group,
            force_recompute=force_recompute,
            cache_handlers=VAEX_DATAFRAME_CACHE_HANDLER,
            disable_cache=disable_cache,
        )

    def _split(self, dataframe: vaex.DataFrame) -> tuple[vaex.DataFrame, vaex.DataFrame]:
        """Split the given dataframe into two parts.

        Args:
            dataframe: The dataframe to be split.

        Returns:
            A tuple containing the split dataframes.
        """
        logger.info(f"Splitting dataframe based on expression {self._expression!r}.")
        filtered_df = dataframe.filter(f"({self._expression})").extract()
        remainder_df = dataframe.filter(f"~({self._expression})").extract()
        logger.info(f"Split dataframe into two dataframes with shapes {filtered_df.shape} and {remainder_df.shape}.")
        return filtered_df, remainder_df



"""Module handling data ingestion from a data source in the pipeline process.

This module contains the `IngestStep` class which is a specialized pipeline step designed for handling data
fetching from a specified `BaseIngester`. It's responsible for retrieving data from the data source, and
returning a `DataContainer` object containing the list of fetched files.
"""

from __future__ import annotations

from typing_extensions import TypedDict

from mleko.dataset.ingest import BaseIngester
from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import PipelineStep
from mleko.utils import auto_repr


class IngestStepInputType(TypedDict):
    """The input type of the IngestStep."""

    pass


class IngestStepOutputType(TypedDict):
    """The output type of the IngestStep."""

    file_paths: str
    """The key for the list of fetched file paths."""


class IngestStep(PipelineStep):
    """Pipeline step that manages data ingestion from a configured data source."""

    _inputs: IngestStepInputType
    _outputs: IngestStepOutputType

    @auto_repr
    def __init__(
        self,
        ingester: BaseIngester,
        inputs: IngestStepInputType,
        outputs: IngestStepOutputType,
    ) -> None:
        """Initialize the IngestStep with the specified data source.

        Args:
            ingester: The data source from which to fetch the data, a BaseIngester instance.
            inputs: A dictionary of input keys following the `IngestStepInputType` schema.
            outputs: A dictionary of output keys following the `IngestStepOutputType` schema.
        """
        super().__init__(inputs, outputs, None)
        self._ingester = ingester

    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Fetch data from the configured data source and return a DataContainer with fetched files.

        Args:
            data_container: Input data for this step's processing operation.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Returns:
            A DataContainer containing the result.
        """
        files = self._ingester.fetch_data(force_recompute)
        data_container.data[self._outputs["file_paths"]] = files
        return data_container

    def _get_input_model(self) -> type[IngestStepInputType]:
        """Get the input type for the IngestStep.

        Returns:
            Input type for the IngestStep.
        """
        return IngestStepInputType

    def _get_output_model(self) -> type[IngestStepOutputType]:
        """Get the output type for the IngestStep.

        Returns:
            Output type for the IngestStep.
        """
        return IngestStepOutputType



"""Module containing the SplitStep class."""

from __future__ import annotations

from typing import Union

from typing_extensions import TypedDict
from vaex import DataFrame

from mleko.dataset.split.base_splitter import BaseSplitter
from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import PipelineStep
from mleko.utils.decorators import auto_repr


class SplitStepInputType(TypedDict):
    """The input type of the SplitStep."""

    dataframe: Union[str, DataFrame]
    """DataFrame or the key for the DataFrame to be split."""


class SplitStepOutputType(TypedDict):
    """The output type of the SplitStep."""

    dataframe_1: str
    """The key for the first DataFrame after splitting."""

    dataframe_2: str
    """The key for the second DataFrame after splitting."""


class SplitStep(PipelineStep):
    """Pipeline step that splits a DataFrame into two parts."""

    _inputs: SplitStepInputType
    _outputs: SplitStepOutputType

    @auto_repr
    def __init__(
        self,
        splitter: BaseSplitter,
        inputs: SplitStepInputType,
        outputs: SplitStepOutputType,
        cache_group: str | None = None,
    ) -> None:
        """Initialize the SplitStep with the specified data splitter.

        Args:
            splitter: The DataSplitter responsible for handling data splitting.
            inputs: A dictionary of input keys following the `SplitStepInputType` schema.
            outputs: A dictionary of output keys following the `SplitStepOutputType` schema.
            cache_group: The cache group to use.
        """
        super().__init__(inputs, outputs, cache_group)
        self._splitter = splitter

    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Perform data splitting using the configured splitter.

        Args:
            data_container: Contains the DataFrame to be split.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Returns:
            A DataContainer containing the result.
        """
        dataframe = self._validate_and_get_input(self._inputs["dataframe"], DataFrame, data_container)

        df1, df2 = self._splitter.split(dataframe, self._cache_group, force_recompute)
        data_container.data[self._outputs["dataframe_1"]] = df1
        data_container.data[self._outputs["dataframe_2"]] = df2
        return data_container

    def _get_input_model(self) -> type[SplitStepInputType]:
        """Get the input type for the SplitStep.

        Returns:
            Input type for the SplitStep.
        """
        return SplitStepInputType

    def _get_output_model(self) -> type[SplitStepOutputType]:
        """Get the output type for the SplitStep.

        Returns:
            Output type for the SplitStep.
        """
        return SplitStepOutputType



"""The subpackage provides common cache handlers for serializing and deserializing data to the disk.

To implement a custom cache handler, you need to create a `CacheHandler` object with the following attributes:
    - `writer`: A function that takes a Path object and data as input and saves the data to the cache file.
    - `reader`: A function that takes a Path object as input and returns the deserialized data stored in the cache file.
    - `suffix`: The suffix of the cache files.

The following cache handlers are provided by the subpackage:
    - `JOBLIB_CACHE_HANDLER`: A cache handler for Python objects using joblib
    - `PICKLE_CACHE_HANDLER`: A cache handler for pickling Python objects.
    - `VAEX_DATAFRAME_CACHE_HANDLER`: A cache handler for `vaex` DataFrames.
"""

from .base_cache_handler import CacheHandler
from .joblib_cache_handler import JOBLIB_CACHE_HANDLER, read_joblib, write_joblib
from .pickle_cache_handler import PICKLE_CACHE_HANDLER, read_pickle, write_pickle
from .vaex_cache_handler import VAEX_DATAFRAME_CACHE_HANDLER, read_vaex_dataframe, write_vaex_dataframe


__all__ = [
    "CacheHandler",
    "JOBLIB_CACHE_HANDLER",
    "PICKLE_CACHE_HANDLER",
    "VAEX_DATAFRAME_CACHE_HANDLER",
    "read_joblib",
    "write_joblib",
    "read_pickle",
    "write_pickle",
    "read_vaex_dataframe",
    "write_vaex_dataframe",
]



"""The module contains a fingerprinter for `Optuna` samplers."""

from __future__ import annotations

import hashlib
import inspect
import json
from typing import Any, Callable

import optuna
from optuna.samplers._lazy_random_state import LazyRandomState
from optuna.samplers._nsgaiii._elite_population_selection_strategy import NSGAIIIElitePopulationSelectionStrategy
from optuna.samplers.nsgaii._after_trial_strategy import NSGAIIAfterTrialStrategy
from optuna.samplers.nsgaii._child_generation_strategy import NSGAIIChildGenerationStrategy
from optuna.samplers.nsgaii._crossovers._blxalpha import BLXAlphaCrossover
from optuna.samplers.nsgaii._crossovers._sbx import SBXCrossover
from optuna.samplers.nsgaii._crossovers._spx import SPXCrossover
from optuna.samplers.nsgaii._crossovers._undx import UNDXCrossover
from optuna.samplers.nsgaii._crossovers._uniform import UniformCrossover
from optuna.samplers.nsgaii._crossovers._vsbx import VSBXCrossover
from optuna.samplers.nsgaii._elite_population_selection_strategy import NSGAIIElitePopulationSelectionStrategy

from mleko.utils.custom_logger import CustomLogger

from .base_fingerprinter import BaseFingerprinter


logger = CustomLogger()
"""The logger for the module."""


class OptunaSamplerFingerprinter(BaseFingerprinter):
    """Class to generate unique fingerprints for different types of Optuna samplers."""

    def fingerprint(self, data: optuna.samplers.BaseSampler) -> str:
        """Generate a fingerprint string for a given Optuna sampler.

        Args:
            data: The sampler to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the sampler's configuration.
        """
        class_name = data.__class__.__name__.lower()
        method_name = f"_fingerprint_{class_name}"

        if hasattr(self, method_name):
            fingerprint_method = getattr(self, method_name)
            return fingerprint_method(data)
        else:
            logger.warning(
                f"Cannot fingerprint sampler of type {class_name}, ensure you invalidate the cache manually."
            )
            return class_name

    def _fingerprint_gridsampler(self, sampler: optuna.samplers.GridSampler) -> str:
        """Generate a fingerprint string for an Optuna GridSampler.

        Args:
            sampler: The GridSampler to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the GridSampler's configuration.
        """
        fingerprint = ""
        fingerprint += self._get_sorted_json_dump(sampler._search_space)
        fingerprint += self._get_rng_state(sampler._rng)
        return hashlib.md5(fingerprint.encode()).hexdigest()

    def _fingerprint_randomsampler(self, sampler: optuna.samplers.RandomSampler) -> str:
        """Generate a fingerprint string for an Optuna RandomSampler.

        Args:
            sampler: The RandomSampler to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the RandomSampler's configuration.
        """
        fingerprint = ""
        fingerprint += self._get_rng_state(sampler._rng)
        return hashlib.md5(fingerprint.encode()).hexdigest()

    def _fingerprint_tpesampler(self, sampler: optuna.samplers.TPESampler) -> str:
        """Generate a fingerprint string for an Optuna TPESampler.

        Args:
            sampler: The TPESampler to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the TPESampler's configuration.
        """
        fingerprint = ""
        fingerprint += str(sampler._parzen_estimator_parameters.consider_prior)
        fingerprint += str(sampler._parzen_estimator_parameters.prior_weight)
        fingerprint += str(sampler._parzen_estimator_parameters.consider_magic_clip)
        fingerprint += str(sampler._parzen_estimator_parameters.consider_endpoints)
        fingerprint += str(sampler._parzen_estimator_parameters.multivariate)
        fingerprint += self._get_inspect_source(sampler._parzen_estimator_parameters.weights)
        fingerprint += str(sampler._n_startup_trials)
        fingerprint += str(sampler._n_ei_candidates)
        fingerprint += str(sampler._gamma)
        fingerprint += self._get_rng_state(sampler._rng)
        fingerprint += self.fingerprint(sampler._random_sampler)
        fingerprint += str(sampler._multivariate)
        fingerprint += str(sampler._group)
        fingerprint += str(sampler._constant_liar)
        fingerprint += self._get_inspect_source(sampler._constraints_func)
        return hashlib.md5(fingerprint.encode()).hexdigest()

    def _fingerprint_cmaessampler(self, sampler: optuna.samplers.CmaEsSampler) -> str:
        """Generate a fingerprint string for an Optuna CmaEsSampler.

        Args:
            sampler: The CmaEsSampler to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the CmaEsSampler's configuration.
        """
        fingerprint = ""
        fingerprint += self._get_sorted_json_dump(sampler._x0)
        fingerprint += str(sampler._sigma0)
        fingerprint += self.fingerprint(sampler._independent_sampler)
        fingerprint += str(sampler._n_startup_trials)
        fingerprint += self._get_rng_state(sampler._cma_rng)
        fingerprint += str(sampler._consider_pruned_trials)
        fingerprint += str(sampler._restart_strategy)
        fingerprint += str(sampler._initial_popsize)
        fingerprint += str(sampler._inc_popsize)
        fingerprint += str(sampler._use_separable_cma)
        fingerprint += str(sampler._with_margin)
        fingerprint += str(sampler._lr_adapt)
        fingerprint += str(sampler._source_trials)
        return hashlib.md5(fingerprint.encode()).hexdigest()

    def _fingerprint_partialfixedsampler(self, sampler: optuna.samplers.PartialFixedSampler) -> str:
        """Generate a fingerprint string for an Optuna PartialFixedSampler.

        Args:
            sampler: The PartialFixedSampler to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the PartialFixedSampler's configuration.
        """
        fingerprint = ""
        fingerprint += self._get_sorted_json_dump(sampler._fixed_params)
        fingerprint += self.fingerprint(sampler._base_sampler)
        return hashlib.md5(fingerprint.encode()).hexdigest()

    def _fingerprint_nsgaiisampler(self, sampler: optuna.samplers.NSGAIISampler) -> str:
        """Generate a fingerprint string for an Optuna NSGAIISampler.

        Args:
            sampler: The NSGAIISampler to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the NSGAIISampler's configuration.
        """
        fingerprint = ""
        fingerprint += str(sampler._population_size)
        if isinstance(sampler._child_generation_strategy, NSGAIIChildGenerationStrategy):
            fingerprint += self._get_nsgaiichildgenerationstrategy(sampler._child_generation_strategy)
        else:
            fingerprint += self._get_inspect_source(sampler._child_generation_strategy)
        fingerprint += self.fingerprint(sampler._random_sampler)
        fingerprint += self._get_rng_state(sampler._rng)
        fingerprint += self._get_inspect_source(sampler._constraints_func)
        if isinstance(sampler._elite_population_selection_strategy, NSGAIIElitePopulationSelectionStrategy):
            fingerprint += str(sampler._elite_population_selection_strategy._population_size)
            fingerprint += self._get_inspect_source(sampler._elite_population_selection_strategy._constraints_func)
        else:
            fingerprint += self._get_inspect_source(sampler._elite_population_selection_strategy)
        if isinstance(sampler._after_trial_strategy, NSGAIIAfterTrialStrategy):
            fingerprint += self._get_inspect_source(sampler._after_trial_strategy._constraints_func)
        else:
            fingerprint += self._get_inspect_source(sampler._after_trial_strategy)
        return hashlib.md5(fingerprint.encode()).hexdigest()

    def _fingerprint_nsgaiiisampler(self, sampler: optuna.samplers.NSGAIIISampler) -> str:
        """Generate a fingerprint string for an Optuna NSGAIIISampler.

        Args:
            sampler: The NSGAIIISampler to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the NSGAIIISampler's configuration.
        """
        fingerprint = ""
        fingerprint += str(sampler._population_size)
        fingerprint += self.fingerprint(sampler._random_sampler)
        fingerprint += self._get_rng_state(sampler._rng)
        fingerprint += self._get_inspect_source(sampler._constraints_func)

        if isinstance(sampler._elite_population_selection_strategy, NSGAIIIElitePopulationSelectionStrategy):
            fingerprint += str(sampler._elite_population_selection_strategy._population_size)
            fingerprint += str(sampler._elite_population_selection_strategy._reference_points)
            fingerprint += str(sampler._elite_population_selection_strategy._dividing_parameter)
            fingerprint += self._get_rng_state(sampler._elite_population_selection_strategy._rng)
            fingerprint += self._get_inspect_source(sampler._elite_population_selection_strategy._constraints_func)
        else:
            fingerprint += self._get_inspect_source(sampler._elite_population_selection_strategy)
        if isinstance(sampler._child_generation_strategy, NSGAIIChildGenerationStrategy):
            fingerprint += self._get_nsgaiichildgenerationstrategy(sampler._child_generation_strategy)
        else:
            fingerprint += self._get_inspect_source(sampler._child_generation_strategy)
        if isinstance(sampler._after_trial_strategy, NSGAIIAfterTrialStrategy):
            fingerprint += self._get_inspect_source(sampler._after_trial_strategy._constraints_func)
        else:
            fingerprint += self._get_inspect_source(sampler._after_trial_strategy)
        return hashlib.md5(fingerprint.encode()).hexdigest()

    def _fingerprint_qmcsampler(self, sampler: optuna.samplers.QMCSampler) -> str:
        """Generate a fingerprint string for an Optuna QMCSampler.

        Args:
            sampler: The QMCSampler to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the QMCSampler's configuration.
        """
        fingerprint = ""
        fingerprint += str(sampler._qmc_type)
        fingerprint += str(sampler._scramble)
        fingerprint += str(sampler._seed)
        fingerprint += self.fingerprint(sampler._independent_sampler)
        return hashlib.md5(fingerprint.encode()).hexdigest()

    def _fingerprint_bruteforcesampler(self, sampler: optuna.samplers.BruteForceSampler) -> str:
        """Generate a fingerprint string for an Optuna BruteForceSampler.

        Args:
            sampler: The BruteForceSampler to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the BruteForceSampler's configuration.
        """
        fingerprint = ""
        fingerprint += self._get_rng_state(sampler._rng)
        return hashlib.md5(fingerprint.encode()).hexdigest()

    def _get_nsgaiichildgenerationstrategy(self, strategy: NSGAIIChildGenerationStrategy) -> str:
        """Generate a fingerprint string for an Optuna NSGAIIChildGenerationStrategy.

        Args:
            strategy: The NSGAIIChildGenerationStrategy to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the NSGAIIChildGenerationStrategy's configuration.
        """
        fingerprint = ""
        fingerprint += str(strategy._crossover_prob)
        fingerprint += str(strategy._mutation_prob)
        fingerprint += str(strategy._swapping_prob)

        fingerprint += str(strategy._crossover.__class__.__name__)
        if isinstance(strategy._crossover, BLXAlphaCrossover):
            fingerprint += str(strategy._crossover._alpha)
        elif isinstance(strategy._crossover, SBXCrossover):
            fingerprint += str(strategy._crossover._eta)
        elif isinstance(strategy._crossover, SPXCrossover):
            fingerprint += str(strategy._crossover._epsilon)
        elif isinstance(strategy._crossover, UNDXCrossover):
            fingerprint += str(strategy._crossover._sigma_xi)
            fingerprint += str(strategy._crossover._sigma_eta)
        elif isinstance(strategy._crossover, UniformCrossover):
            fingerprint += str(strategy._crossover._swapping_prob)
        elif isinstance(strategy._crossover, VSBXCrossover):
            fingerprint += str(strategy._crossover._eta)

        fingerprint += self._get_inspect_source(strategy._constraints_func)
        fingerprint += self._get_rng_state(strategy._rng)
        return fingerprint

    def _get_inspect_source(self, func: Callable | None) -> str:
        """Get the source code of a Callable.

        Args:
            func: The Callable to get the source code of.

        Returns:
            The source code of the Callable.
        """
        return inspect.getsource(func) if func is not None else ""

    def _get_sorted_json_dump(self, data: dict[str, Any] | None) -> str:
        """Get a sorted JSON dump of a dictionary.

        Args:
            data: The dictionary to dump.

        Returns:
            The sorted JSON dump of the dictionary.
        """

        def deep_sort(obj: dict | Any):
            """Recursively sort nested dicts.

            Args:
                obj: The dict or object to sort.

            Returns:
                Sorted dict, or object.
            """
            if isinstance(obj, dict):
                _sorted = {}
                for key in sorted(obj):
                    _sorted[key] = deep_sort(obj[key])
                return _sorted
            else:
                return obj

        sorted_data = deep_sort(data)
        return json.dumps(sorted_data, sort_keys=True) if data is not None else ""

    def _get_rng_state(self, rng: LazyRandomState) -> str:
        """Get the state of a RandomState.

        Args:
            rng: The RandomState to get the state of.

        Returns:
            The state of the RandomState.
        """
        return str(rng.rng.get_state())



"""The module containing the base class for data filter."""

from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path

import vaex

from mleko.cache.lru_cache_mixin import LRUCacheMixin
from mleko.dataset.data_schema import DataSchema


class BaseFilter(LRUCacheMixin, ABC):
    """Abstract base class for data filter.

    Will cache the filtered dataframes in the output directory.
    """

    def __init__(self, cache_directory: str | Path, cache_size: int) -> None:
        """Initializes the `BaseFilter` with an output directory.

        Args:
            cache_directory: The target directory where the filtered dataframes are to be saved.
            cache_size: The maximum number of cache entries.
        """
        super().__init__(cache_directory, cache_size)

    @abstractmethod
    def filter(
        self,
        data_schema: DataSchema,
        dataframe: vaex.DataFrame,
        cache_group: str | None = None,
        force_recompute: bool = False,
        disable_cache: bool = False,
    ) -> vaex.DataFrame:
        """Abstract method to filter the given dataframe.

        Args:
            data_schema: The data schema to be used for filtering.
            dataframe: The dataframe to be filtered.
            cache_group: The cache group to use.
            force_recompute: Forces recomputation if True, otherwise reads from the cache if available.
            disable_cache: If set to True, disables the cache.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError



"""This module contains the `CacheHandler` class."""

from pathlib import Path
from typing import Any, Callable, NamedTuple


class CacheHandler(NamedTuple):
    """A named tuple representing a cache handler."""

    writer: Callable[[Path, Any], None]
    """A function that takes a Path object and data as input and saves the data to the cache file."""

    reader: Callable[[Path], Any]
    """A function that takes a Path object as input and returns the deserialized data stored in the cache file."""

    suffix: str
    """The suffix of the cache files."""

    can_handle_none: bool
    """Whether the cache handler can handle None values."""



"""The subpackage provides functionality for filtering `vaex` DataFrames.

The subpackage contains the following filter classes:
    - `BaseFilter`: The abstract base class for all filters.
    - `ExpressionFilter`: A class that handles filtering of `vaex` DataFrames based on a given `vaex` expression.
"""

from __future__ import annotations

from .base_filter import BaseFilter
from .expression_filter import ExpressionFilter
from .imblearn_resampling_filter import ImblearnResamplingFilter


__all__ = ["BaseFilter", "ExpressionFilter", "ImblearnResamplingFilter"]



"""Module for the variance feature selector."""

from __future__ import annotations

from pathlib import Path
from typing import Hashable

import vaex
from tqdm.auto import tqdm
from vaex.ml import MaxAbsScaler

from mleko.dataset.data_schema import DataSchema
from mleko.dataset.feature_select.base_feature_selector import BaseFeatureSelector
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr
from mleko.utils.vaex_helpers import get_column, get_columns


logger = CustomLogger()
"""A module-level logger for the module."""


class VarianceFeatureSelector(BaseFeatureSelector):
    """Selects features based on the variance."""

    @auto_repr
    def __init__(
        self,
        variance_threshold: float,
        features: list[str] | tuple[str, ...] | None = None,
        ignore_features: list[str] | tuple[str, ...] | None = None,
        cache_directory: str | Path = "data/variance-feature-selector",
        cache_size: int = 1,
    ) -> None:
        """Initializes the feature selector.

        The feature selector will select all features with a variance above the specified threshold.
        The default set of features is all numeric features in the DataFrame.

        Note:
            Only works with numeric features.

        Warning:
            Make sure to ignore any important features that need to be kept, such as the
            target feature or some identifier.

        Args:
            variance_threshold: The minimum variance allowed for a feature to be selected.
            features: List of feature names to be used by the feature selector.
            ignore_features: List of feature names to be ignored by the feature selector.
            cache_directory: Directory where the cache will be stored locally.
            cache_size: The maximum number of entries to keep in the cache.

        Examples:
            >>> import vaex
            >>> from mleko.dataset.feature_select import VarianceFeatureSelector
            >>> from mleko.utils.vaex_helpers import get_column
            >>> df = vaex.from_arrays(
            ...     a=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            ...     b=[1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            ...     c=[1, 2, 2, 2, 2, 2, 2, 2, 2, 2],
            ...     d=["str1", "str2", "str3", "str4", "str5", "str6", "str7", "str8", "str9", "str10"],
            ... )
            >>> ds = DataSchema(
            ...     numerical=["a", "b", "c"],
            ...     categorical=["d"],
            ... )
            >>> selector = VarianceFeatureSelector(
            ...     variance_threshold=0.1,
            ...     ignore_features=["c"],
            ... )
            >>> ds, _, df = selector.fit_transform(ds, df)
            >>> df.get_column_names()
            ['a', 'c', 'd']
        """
        super().__init__(features, ignore_features, cache_directory, cache_size)
        self._variance_threshold = variance_threshold
        self._feature_selector: set[str] = set()

    def _fit(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, set[str]]:
        """Fits the feature selector on the input data.

        Args:
            data_schema: The DataSchema of the DataFrame.
            dataframe: The DataFrame to fit the feature selector on.

        Returns:
            Updated DataSchema and the set of features with a variance below the threshold.
        """
        features = self._feature_set(data_schema)
        logger.info(f"Fitting variance feature selector on {len(features)} features: {features}.")

        if self._variance_threshold > 0:
            scaler = MaxAbsScaler(features=list(features), prefix="")
            df = scaler.fit_transform(dataframe)
        else:
            df = dataframe

        variance: dict[str, float] = {}
        for feature in tqdm(features, desc="Calculating variance for features"):
            column = get_column(df, feature)
            variance[feature] = column.var()

        self._feature_selector = {feature for feature in features if variance[feature] <= self._variance_threshold}
        ds = data_schema.copy().drop_features(self._feature_selector)

        return ds, self._feature_selector

    def _transform(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, vaex.DataFrame]:
        """Selects features based on the variance.

        Args:
            data_schema: The DataSchema of the DataFrame.
            dataframe: The DataFrame to select features from.

        Returns:
            Updated DataSchema and DataFrame with the selected features.
        """
        dropped_features = self._feature_selector
        logger.info(
            f"Dropping ({len(dropped_features)}) features with normalized variance <= {self._variance_threshold}: "
            f"{dropped_features}."
        )
        selected_features = [feature for feature in dataframe.get_column_names() if feature not in dropped_features]
        ds = data_schema.copy().drop_features(dropped_features)

        return ds, get_columns(dataframe, selected_features)

    def _default_features(self, data_schema: DataSchema) -> tuple[str, ...]:
        """Returns the default set of features.

        Args:
            data_schema: The DataSchema of the DataFrame.

        Returns:
            Tuple of default features.
        """
        features = data_schema.get_features(["numerical"])
        return tuple(str(feature) for feature in features)

    def _fingerprint(self) -> Hashable:
        """Returns a hashable fingerprint of the feature selector.

        Append the variance threshold to the fingerprint.

        Returns:
            The fingerprint of the feature selector.
        """
        return super()._fingerprint(), self._variance_threshold



"""Module for the composite transformer."""

from __future__ import annotations

from pathlib import Path
from typing import Any, Hashable

import vaex

from mleko.dataset.data_schema import DataSchema
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr

from .base_transformer import BaseTransformer


logger = CustomLogger()
"""A module-level logger for the module."""


class CompositeTransformer(BaseTransformer):
    """A transformer that combines multiple transformers.

    It is possible to combine multiple transformers into a single transformer. This can be useful when multiple
    transformers need to be applied to a DataFrame and storing the intermediate DataFrames is not desired.
    """

    @auto_repr
    def __init__(
        self,
        transformers: list[BaseTransformer] | tuple[BaseTransformer, ...],
        cache_directory: str | Path = "data/composite-transformer",
        cache_size: int = 1,
    ) -> None:
        """Initializes the composite transformer.

        The composite transformer will combine the transformers into a single transformer. Each transformer will be
        applied to the DataFrame in the order they are specified. Caching of the intermediate DataFrames is disabled
        and will only be performed on the final DataFrame.

        Args:
            transformers: List of transformers to be combined.
            cache_directory: Directory where the cache will be stored locally.
            cache_size: The maximum number of entries to keep in the cache.

        Examples:
            >>> import vaex
            >>> from mleko.dataset.transform import (
            ...     CompositeTransformer,
            ...     LabelEncoderTransformer,
            ...     FrequencyEncoderTransformer
            ... )
            >>> df = vaex.from_arrays(
            ...     a=["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"],
            ...     b=["a", "a", "a", "a", None, None, None, None, None, None],
            ... )
            >>> ds = DataSchema(
            ...     categorical=["a", "b"],
            ... )
            >>> transformer = CompositeTransformer(
            ...     transformers=[
            ...         LabelEncoderTransformer(
            ...             features=["a"],
            ...         ),
            ...         FrequencyEncoderTransformer(
            ...             features=["b"],
            ...         ),
            ...     ],
            ... )
            >>> _, _, df = transformer.fit_transform(ds, df)
            >>> df["a"].tolist()
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            >>> df["b"].tolist()
            [0.4, 0.4, 0.4, 0.4, nan, nan, nan, nan, nan, nan]
        """
        super().__init__([], cache_directory, cache_size)
        self._transformers = tuple(transformers)
        self._transformer: list[Any] = []

    def _fit(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, list[Any]]:
        """Fits the transformer to the specified DataFrame.

        Args:
            data_schema: Data schema of the DataFrame.
            dataframe: DataFrame to be fitted.

        Returns:
            Updated data schema and list of fitted transformers.
        """
        fitted_transformers: list[Any] = []
        for i, transformer in enumerate(self._transformers):
            logger.info(
                f"Fitting composite feature transformation step {i+1}/{len(self._transformers)}: "
                f"{transformer.__class__.__name__}."
            )
            data_schema, fitted_transformer = transformer.fit(data_schema, dataframe, disable_cache=True)
            fitted_transformers.append(fitted_transformer)
            logger.info(f"Finished fitting composite transformation step {i+1}/{len(self._transformers)}.")
        return data_schema, fitted_transformers

    def _transform(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, vaex.DataFrame]:
        """Returns the updated data schema transformed DataFrame.

        Args:
            data_schema: The data schema of the DataFrame.
            dataframe: The DataFrame to transform.

        Returns:
            Updated data schema and transformed DataFrame.
        """
        for i, transformer in enumerate(self._transformers):
            logger.info(
                f"Executing composite feature transformation step {i+1}/{len(self._transformers)}: "
                f"{transformer.__class__.__name__}."
            )
            data_schema, dataframe = transformer.transform(data_schema, dataframe, disable_cache=True)
            dataframe = dataframe.extract()
            logger.info(f"Finished composite transformation step {i+1}/{len(self._transformers)}.")
        return data_schema, dataframe

    def _fit_transform(
        self, data_schema: DataSchema, dataframe: vaex.DataFrame
    ) -> tuple[DataSchema, Any, vaex.DataFrame]:
        """Fits the transformer to the specified DataFrame and performs the transformation on the DataFrame.

        Args:
            data_schema: The data schema of the DataFrame.
            dataframe: The DataFrame to transform.

        Returns:
            Tuple of updated data schema, fitted transformer and transformed DataFrame.
        """
        fitted_transformers: list[Any] = []
        for i, transformer in enumerate(self._transformers):
            logger.info(
                f"Executing composite transformation step {i+1}/{len(self._transformers)}: "
                f"{transformer.__class__.__name__}."
            )
            data_schema, fitted_transformer, dataframe = transformer.fit_transform(
                data_schema, dataframe, disable_cache=True
            )
            fitted_transformers.append(fitted_transformer)
            dataframe = dataframe.extract()
            logger.info(
                "Finished fitting and transforming composite transformation " f"step {i+1}/{len(self._transformers)}."
            )
        return data_schema, fitted_transformers, dataframe

    def _assign_transformer(self, transformer: Any) -> None:
        """Assigns the specified transformer to the transformer attribute.

        Can be overridden by subclasses to assign the transformer using a different method.

        Args:
            transformer: Transformer to be assigned.
        """
        self._transformer = transformer
        for transformer_obj, fitted_transformer in zip(self._transformers, transformer):
            transformer_obj._transformer = fitted_transformer

    def _fingerprint(self) -> Hashable:
        """Returns the fingerprint of the transformer.

        Returns:
            A hashable object that uniquely identifies the transformer.
        """
        return super()._fingerprint(), tuple(transformer._fingerprint() for transformer in self._transformers)



"""Module for the base model class."""

from __future__ import annotations

from abc import ABC, abstractmethod
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, Hashable, Union

import pandas as pd
import vaex

from mleko.cache.fingerprinters import DictFingerprinter, VaexFingerprinter
from mleko.cache.handlers import JOBLIB_CACHE_HANDLER, VAEX_DATAFRAME_CACHE_HANDLER
from mleko.cache.lru_cache_mixin import LRUCacheMixin
from mleko.dataset.data_schema import DataSchema
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.vaex_helpers import HashableVaexDataFrame, get_columns


logger = CustomLogger()
"""The logger for the module."""

HyperparametersType = Dict[
    str,
    Union[
        str,
        int,
        float,
        bool,
        list,
        tuple,
        None,
    ],
]


class BaseModel(LRUCacheMixin, ABC):
    """Abstract class for models.

    The model fitting and transformation process is implemented in the `fit`, `transform`, and `fit_transform` methods,
    similar to the scikit-learn API. The `fit` method fits the model to the specified DataFrame, the `transform`
    method transforms the specified features in the DataFrame, and the `fit_transform` method fits the model to the
    specified DataFrame and transforms the specified features in the DataFrame.
    """

    def __init__(
        self,
        features: list[str] | tuple[str, ...] | None,
        ignore_features: list[str] | tuple[str, ...] | None,
        verbosity: int,
        memoized_dataset_cache_size: int | None,
        cache_directory: str | Path,
        cache_size: int,
    ) -> None:
        """Initializes the model and ensures the destination directory exists.

        Note:
            The `features` and `ignore_features` arguments are mutually exclusive. If both are specified, a
            `ValueError` is raised.

        Warning:
            The `memoized_dataset_cache_size` parameter is experimental and should be used with caution. It refers to
            the number of datasets to keep in memory for speeding up repeated training. This can be useful when
            hyperparameter tuning or cross-validation is performed, as the dataset does not need to be loaded from disk
            every time. However, this can lead to memory issues if the dataset is too large. Specify 0 to disable the
            cache. When finished with the fitting and transforming, please call the `_clear_dataset_cache` method to
            clear the cache and free up memory.

        Args:
            features: List of feature names to be used by the model. If None, the default is all features
                applicable to the model.
            ignore_features: List of feature names to be ignored by the model. If None, the default is to
                ignore no features.
            verbosity: The verbosity level of the logger, should be passed to the underlying model.
            memoized_dataset_cache_size: The number of datasets to keep in memory for speeding up repeated training.
                When finished with the fitting and transforming, please call the `_clear_dataset_cache` method to clear
                the cache and free up memory. Specify 0 to disable the cache.
            cache_directory: Directory where the cache will be stored locally.
            cache_size: The maximum number of entries to keep in the cache.

        Raises:
            ValueError: If both `features` and `ignore_features` are specified.
        """
        super().__init__(cache_directory, cache_size)
        self._memoized_dataset_cache_size = memoized_dataset_cache_size

        if features is not None and ignore_features is not None:
            msg = "Both `features` and `ignore_features` have been specified. The arguments are mutually exclusive."
            logger.error(msg)
            raise ValueError(msg)

        self._model = None
        self._hyperparameters: HyperparametersType = {}
        self._features: tuple[str, ...] | None = tuple(features) if features is not None else None
        self._ignore_features: tuple[str, ...] = tuple(ignore_features) if ignore_features is not None else tuple()
        logger.set_level(verbosity)

        self._memoized_load_dataset = lru_cache(maxsize=self._memoized_dataset_cache_size)(self._memoized_load_dataset)

    def fit(
        self,
        data_schema: DataSchema,
        train_dataframe: vaex.DataFrame,
        validation_dataframe: vaex.DataFrame | None = None,
        hyperparameters: HyperparametersType | None = None,
        cache_group: str | None = None,
        force_recompute: bool = False,
        disable_cache: bool = False,
    ) -> tuple[Any, dict[str, dict[str, list[Any]]]]:
        """Fits the model to the specified DataFrame, using the specified hyperparameters.

        The validation DataFrame is used to validate the model during fitting.

        Args:
            data_schema: Data schema for the DataFrame.
            train_dataframe: DataFrame to fit the model on.
            validation_dataframe: DataFrame to be used for validation.
            hyperparameters: Hyperparameters to be used for fitting. If any hyperparameters are specified, they will
                be merged with the default hyperparameters specified during the model initialization.
            cache_group: The cache group to use for caching.
            force_recompute: Whether to force recompute the result.
            disable_cache: If set to True, disables the cache.

        Returns:
            Fitted model and the metrics dictionary. The metrics dictionary is a dictionary of dictionaries. The outer
            dictionary is keyed by the dataset name, and the inner dictionary is keyed by the metric name. The value
            of the inner dictionary is a list of metric values for each iteration of the model.
            >>> metrics = {
            ...     "train": {
            ...         "accuracy": [0.90, 0.91, 0.92],
            ...         "f1": [0.80, 0.81, 0.82],
            ...     },
            ...     "validation": {
            ...         "accuracy": [0.80, 0.81, 0.82],
            ...         "f1": [0.70, 0.71, 0.72],
            ...     },
            ... }
        """
        model, metrics = self._cached_execute(
            lambda_func=lambda: self._fit(data_schema, train_dataframe, validation_dataframe, hyperparameters),
            cache_key_inputs=[
                self._fingerprint(),
                (
                    (
                        {**self._hyperparameters, **hyperparameters}
                        if hyperparameters is not None
                        else self._hyperparameters
                    ),
                    DictFingerprinter(),
                ),
                (data_schema.to_dict(), DictFingerprinter()),
                (train_dataframe, VaexFingerprinter()),
                (validation_dataframe, VaexFingerprinter()) if validation_dataframe is not None else "None",
            ],
            cache_group=cache_group,
            force_recompute=force_recompute,
            cache_handlers=JOBLIB_CACHE_HANDLER,
            disable_cache=disable_cache,
        )
        self._assign_model(model)
        return model, metrics

    def transform(
        self,
        data_schema: DataSchema,
        dataframe: vaex.DataFrame,
        cache_group: str | None = None,
        force_recompute: bool = False,
        disable_cache: bool = False,
    ) -> vaex.DataFrame:
        """Transforms the specified DataFrame using the fitted model.

        Args:
            data_schema: Data schema for the DataFrame.
            dataframe: DataFrame to be transformed.
            cache_group: The cache group to use for caching.
            force_recompute: Whether to force recompute the result.
            disable_cache: If set to True, disables the cache.

        Raises:
            RuntimeError: If the model has not been fitted.

        Returns:
            Transformed DataFrame.
        """
        if self._model is None:
            msg = "Model must be fitted before it can be used to transform the DataFrame."
            logger.error(msg)
            raise RuntimeError(msg)

        return self._cached_execute(
            lambda_func=lambda: self._transform(data_schema, dataframe),
            cache_key_inputs=[
                self._fingerprint(),
                (self._hyperparameters, DictFingerprinter()),
                (data_schema.to_dict(), DictFingerprinter()),
                (dataframe, VaexFingerprinter()),
            ],
            cache_group=cache_group,
            force_recompute=force_recompute,
            cache_handlers=VAEX_DATAFRAME_CACHE_HANDLER,
            disable_cache=disable_cache,
        )

    def fit_transform(
        self,
        data_schema: DataSchema,
        train_dataframe: vaex.DataFrame,
        validation_dataframe: vaex.DataFrame | None = None,
        hyperparameters: HyperparametersType | None = None,
        cache_group: str | None = None,
        force_recompute: bool = False,
        disable_cache: bool = False,
    ) -> tuple[Any, dict[str, dict[str, list[Any]]], vaex.DataFrame, vaex.DataFrame | None]:
        """Fits the model to the specified DataFrame and transforms the train and validation DataFrames.

        The validation DataFrame is used to validate the model during fitting.

        Args:
            data_schema: Data schema for the DataFrame.
            train_dataframe: DataFrame to fit the model on.
            validation_dataframe: DataFrame to be used for validation.
            hyperparameters: Hyperparameters to be used for fitting.
            cache_group: The cache group to use for caching.
            force_recompute: Whether to force recompute the result.
            disable_cache: If set to True, disables the cache.

        Returns:
            Tuple of fitted model, the metrics dictionary, transformed train DataFrame,
            and transformed validation DataFrame. The metrics dictionary is a dictionary of dictionaries.
            The outer dictionary is keyed by the dataset name, and the inner dictionary is keyed by the
            metric name. The value of the inner dictionary is a list of metric values for each
            iteration of the model.
            >>> metrics = {
            ...     "train": {
            ...         "accuracy": [0.90, 0.91, 0.92],
            ...         "f1": [0.80, 0.81, 0.82],
            ...     },
            ...     "validation": {
            ...         "accuracy": [0.80, 0.81, 0.82],
            ...         "f1": [0.70, 0.71, 0.72],
            ...     },
            ... }
        """
        model, metrics, df_train, df_validation = self._cached_execute(
            lambda_func=lambda: self._fit_transform(
                data_schema, train_dataframe, validation_dataframe, hyperparameters
            ),
            cache_key_inputs=[
                self._fingerprint(),
                (
                    (
                        {**self._hyperparameters, **hyperparameters}
                        if hyperparameters is not None
                        else self._hyperparameters
                    ),
                    DictFingerprinter(),
                ),
                (data_schema.to_dict(), DictFingerprinter()),
                (train_dataframe, VaexFingerprinter()),
                (validation_dataframe, VaexFingerprinter()) if validation_dataframe is not None else None,
            ],
            cache_group=cache_group,
            force_recompute=force_recompute,
            cache_handlers=[
                JOBLIB_CACHE_HANDLER,
                JOBLIB_CACHE_HANDLER,
                VAEX_DATAFRAME_CACHE_HANDLER,
                VAEX_DATAFRAME_CACHE_HANDLER,
            ],
            disable_cache=disable_cache,
        )
        self._assign_model(model)
        return model, metrics, df_train, df_validation

    def clear_load_dataset_cache(self) -> None:
        """Clears the cache for the `_memoized_load_dataset` method."""
        self._memoized_load_dataset.cache_clear()  # type: ignore

    def _fit_transform(
        self,
        data_schema: DataSchema,
        train_dataframe: vaex.DataFrame,
        validation_dataframe: vaex.DataFrame | None = None,
        hyperparameters: HyperparametersType | None = None,
    ) -> tuple[Any, dict[str, dict[str, list[Any]]], vaex.DataFrame, vaex.DataFrame | None]:
        """Fits the model to the specified DataFrame and transforms the train and validation DataFrames.

        Args:
            data_schema: Data schema for the DataFrame.
            train_dataframe: DataFrame to fit the model on.
            validation_dataframe: DataFrame to be used for validation.
            hyperparameters: Hyperparameters to be used for fitting.

        Returns:
            Tuple of fitted model, the metrics dictionary, transformed train DataFrame, and
            transformed validation DataFrame.
        """
        model, metrics = self._fit(data_schema, train_dataframe, validation_dataframe, hyperparameters)
        return (
            model,
            metrics,
            self._transform(data_schema, train_dataframe),
            self._transform(data_schema, validation_dataframe) if validation_dataframe is not None else None,
        )

    def _assign_model(self, model: Any) -> None:
        """Assigns the specified model to the model attribute.

        Can be overridden by subclasses to assign the model using a different method.

        Args:
            model: Model to be assigned.
        """
        self._model = model

    def _feature_set(self, data_schema: DataSchema) -> list[str]:
        """Returns the list of features to be used as input by the model.

        It is the default set of features minus the features to be ignored if the `features` argument is None, or the
        list of names in the `features` argument if it is not None.

        Args:
            data_schema: Data schema for the DataFrame.

        Returns:
            Sorted list of feature names to be used by the model.
        """
        return sorted(
            set(self._default_features(data_schema)) - set(self._ignore_features)
            if self._features is None
            else self._features
        )

    def _load_dataset(
        self,
        data_schema: DataSchema,
        dataframe: vaex.DataFrame,
        additional_features: list[str] | None = None,
    ) -> pd.DataFrame:
        """Load the dataset into memory.

        Warning:
            This method should be used with caution, as it loads the entire dataset into memory as a pandas DataFrame.

        Args:
            data_schema: The data schema of the dataframe.
            dataframe: The dataframe to load.
            additional_features: Additional features to load, such as the target feature.

        Returns:
            A pandas DataFrame with the loaded data.
        """
        feature_names = self._feature_set(data_schema)
        df = get_columns(dataframe, feature_names + (additional_features if additional_features else [])).to_pandas_df()
        return df  # type: ignore

    def _memoized_load_dataset(
        self,
        data_schema: DataSchema,
        dataframe: HashableVaexDataFrame,
        additional_features: tuple[str, ...] | None = None,
        name: str | None = None,
    ) -> pd.DataFrame:
        """Load the dataset into memory and memoize the result.

        Warning:
            This method should be used with caution, as it loads the entire dataset into memory as a pandas DataFrame.
            The returned DataFrame will be memoized using the `functools.lru_cache` to avoid reloading the
            dataset multiple times. The cache size is set to the `memoized_dataset_cache_size` attribute.

        Args:
            data_schema: The data schema of the dataframe.
            dataframe: The dataframe to load, wrapped in a `HashableVaexDataFrame` object.
            additional_features: Additional features to load, such as the target feature.
            name: Name of the dataset to be used in the log message.

        Returns:
            A pandas DataFrame with the loaded data.
        """
        if name is not None:
            name = f"{name.strip()} "
        else:
            name = ""

        logger.info(f"Loading the {name}dataset into memory.")
        return self._load_dataset(data_schema, dataframe.df, list(additional_features) if additional_features else None)

    @abstractmethod
    def _fit(
        self,
        data_schema: DataSchema,
        train_dataframe: vaex.DataFrame,
        validation_dataframe: vaex.DataFrame | None = None,
        hyperparameters: HyperparametersType | None = None,
    ) -> tuple[Any, dict[str, dict[str, list[Any]]]]:
        """Fits the model to the specified DataFrame.

        Args:
            data_schema: Data schema for the DataFrame.
            train_dataframe: DataFrame to be fitted.
            validation_dataframe: DataFrame to be used for validation.
            hyperparameters: Hyperparameters to be used for fitting.

        Raises:
            NotImplementedError: Must be implemented by subclasses.
        """
        raise NotImplementedError

    @abstractmethod
    def _transform(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> vaex.DataFrame:
        """Transforms the specified DataFrame using the fitted model.

        Args:
            data_schema: Data schema for the DataFrame.
            dataframe: DataFrame to be transformed.

        Raises:
            NotImplementedError: Must be implemented by subclasses.
        """
        raise NotImplementedError

    @abstractmethod
    def _fingerprint(self) -> Hashable:
        """Returns a hashable object that uniquely identifies the model.

        The base implementation fingerprints the class name and the important attributes of the model.

        Note:
            Subclasses should call the parent method and include the result in the hashable object along with any
            other parameters that uniquely identify the model. All attributes that are used in the
            model that affect the result of the fitting and transforming should be included in the hashable object.

        Returns:
            Hashable object that uniquely identifies the model.
        """
        return (
            self.__class__.__name__,
            self._features,
            self._ignore_features,
        )

    @abstractmethod
    def _default_features(self, data_schema: DataSchema) -> tuple[str, ...]:
        """Returns the default set of features to be used by the model.

        Args:
            data_schema: Data schema for the DataFrame.

        Raises:
            NotImplementedError: Must be implemented in the child class that inherits from `BaseModel`.
        """
        raise NotImplementedError



"""Module for the base feature selector class."""

from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Hashable

import vaex

from mleko.cache.fingerprinters import DictFingerprinter, VaexFingerprinter
from mleko.cache.handlers import JOBLIB_CACHE_HANDLER, VAEX_DATAFRAME_CACHE_HANDLER
from mleko.cache.lru_cache_mixin import LRUCacheMixin
from mleko.dataset.data_schema import DataSchema
from mleko.utils.custom_logger import CustomLogger


logger = CustomLogger()
"""The logger for the module."""


class BaseFeatureSelector(LRUCacheMixin, ABC):
    """Abstract class for feature selection.

    The feature selection process is implemented in the `fit`, `transform`, and `fit_transform` methods, following the
    scikit-learn API. The `fit` method takes a DataFrame as input and returns a fitted feature selector. The `transform`
    method takes a DataFrame as input and returns a transformed DataFrame. The `fit_transform` method takes a DataFrame
    as input and returns a tuple of a fitted feature selector and a transformed DataFrame.

    Note:
        The default set of features to be used by the feature selector is all features applicable to the feature
        selector. This can be overridden by passing a list of feature names to the `features` parameter of the
        constructor. The default set of features to be ignored by the feature selector is no features. This can be
        overridden by passing a list of feature names to the `ignore_features` parameter of the constructor.
    """

    def __init__(
        self,
        features: list[str] | tuple[str, ...] | None,
        ignore_features: list[str] | tuple[str, ...] | None,
        cache_directory: str | Path,
        cache_size: int,
    ) -> None:
        """Initializes the feature selector and ensures the destination directory exists.

        Note:
            The `features` and `ignore_features` arguments are mutually exclusive. If both are specified, a
            `ValueError` is raised.

        Args:
            features: List of feature names to be used by the feature selector. If None, the default is all features
                applicable to the feature selector.
            ignore_features: List of feature names to be ignored by the feature selector. If None, the default is to
                ignore no features.
            cache_directory: Directory where the cache will be stored locally.
            cache_size: The maximum number of cache entries.

        Raises:
            ValueError: If both `features` and `ignore_features` are specified.
        """
        super().__init__(cache_directory, cache_size)
        if features is not None and ignore_features is not None:
            msg = "Both `features` and `ignore_features` have been specified. The arguments are mutually exclusive."
            logger.error(msg)
            raise ValueError(msg)

        self._features: tuple[str, ...] | None = tuple(features) if features is not None else None
        self._ignore_features: tuple[str, ...] = tuple(ignore_features) if ignore_features is not None else tuple()
        self._feature_selector = None

    def fit(
        self,
        data_schema: DataSchema,
        dataframe: vaex.DataFrame,
        cache_group: str | None = None,
        force_recompute: bool = False,
        disable_cache: bool = False,
    ) -> tuple[DataSchema, Any]:
        """Fits the feature selector to the specified DataFrame, using the cached result if available.

        Args:
            data_schema: DataSchema of the DataFrame.
            dataframe: DataFrame to be fitted.
            cache_group: The cache group to use.
            force_recompute: Whether to force the fitting to be recomputed even if the result is cached.
            disable_cache: If set to True, disables the cache.

        Returns:
            Updated DataSchema and fitted feature selector.
        """
        ds, feature_selector = self._cached_execute(
            lambda_func=lambda: self._fit(data_schema, dataframe),
            cache_key_inputs=[
                self._fingerprint(),
                (data_schema.to_dict(), DictFingerprinter()),
                (dataframe, VaexFingerprinter()),
            ],
            cache_group=cache_group,
            force_recompute=force_recompute,
            cache_handlers=JOBLIB_CACHE_HANDLER,
            disable_cache=disable_cache,
        )
        self._assign_feature_selector(feature_selector)
        return ds, feature_selector

    def transform(
        self,
        data_schema: DataSchema,
        dataframe: vaex.DataFrame,
        cache_group: str | None = None,
        force_recompute: bool = False,
        disable_cache: bool = False,
    ) -> tuple[DataSchema, vaex.DataFrame]:
        """Extracts the selected features from the DataFrame, using the cached result if available.

        Args:
            data_schema: DataSchema of the DataFrame.
            dataframe: DataFrame to be transformed.
            cache_group: The cache group to use.
            force_recompute: Whether to force the transformation to be recomputed even if the result is cached.
            disable_cache: If set to True, disables the cache.

        Raises:
            RuntimeError: If the feature selector has not been fitted.

        Returns:
            Updated DataSchema and transformed DataFrame.
        """
        if self._feature_selector is None:
            msg = "Feature selector must be fitted before it can be used to extract selected features."
            logger.error(msg)
            raise RuntimeError(msg)

        ds, df = self._cached_execute(
            lambda_func=lambda: self._transform(data_schema, dataframe),
            cache_key_inputs=[
                self._fingerprint(),
                (data_schema.to_dict(), DictFingerprinter()),
                (dataframe, VaexFingerprinter()),
            ],
            cache_group=cache_group,
            force_recompute=force_recompute,
            cache_handlers=[JOBLIB_CACHE_HANDLER, VAEX_DATAFRAME_CACHE_HANDLER],
            disable_cache=disable_cache,
        )
        return ds, df

    def fit_transform(
        self,
        data_schema: DataSchema,
        dataframe: vaex.DataFrame,
        cache_group: str | None = None,
        force_recompute: bool = False,
        disable_cache: bool = False,
    ) -> tuple[DataSchema, Any, vaex.DataFrame]:
        """Fits the feature selector to the specified DataFrame and extracts the selected features from the DataFrame.

        Args:
            data_schema: DataSchema of the DataFrame.
            dataframe: DataFrame to be fitted and transformed.
            cache_group: The cache group to use.
            force_recompute: Whether to force the fitting and transformation to be recomputed even if the result is
                cached.
            disable_cache: If set to True, disables the cache.

        Returns:
            Tuple of updated DataSchema, fitted feature selector, and transformed DataFrame.
        """
        ds, feature_selector, df = self._cached_execute(
            lambda_func=lambda: self._fit_transform(data_schema, dataframe),
            cache_key_inputs=[
                self._fingerprint(),
                (data_schema.to_dict(), DictFingerprinter()),
                (dataframe, VaexFingerprinter()),
            ],
            cache_group=cache_group,
            force_recompute=force_recompute,
            cache_handlers=[JOBLIB_CACHE_HANDLER, JOBLIB_CACHE_HANDLER, VAEX_DATAFRAME_CACHE_HANDLER],
            disable_cache=disable_cache,
        )
        self._assign_feature_selector(feature_selector)
        return ds, feature_selector, df

    def _fit_transform(
        self, data_schema: DataSchema, dataframe: vaex.DataFrame
    ) -> tuple[DataSchema, Any, vaex.DataFrame]:
        """Fits the feature selector to the specified DataFrame and extracts the selected features from the DataFrame.

        Args:
            data_schema: DataSchema of the DataFrame.
            dataframe: DataFrame used for feature selection.

        Returns:
            Tuple of updated DataSchema, fitted feature selector, and transformed DataFrame.
        """
        ds, feature_selector = self._fit(data_schema, dataframe)
        ds, df = self._transform(data_schema, dataframe)
        return ds, feature_selector, df

    def _assign_feature_selector(self, feature_selector: Any) -> None:
        """Assigns the specified feature selector to the feature_selector attribute.

        Can be overridden by subclasses to assign the feature selector using a different method.

        Args:
            feature_selector: Feature selector to be assigned.
        """
        self._feature_selector = feature_selector

    def _feature_set(self, data_schema: DataSchema) -> list[str]:
        """Returns the list of features to be used by the feature selector.

        It is the default set of features minus the features to be ignored if the `features` argument is None, or the
        list of names in the `features` argument if it is not None.

        Args:
            data_schema: DataSchema of the DataFrame.

        Returns:
            Sorted list of feature names to be used by the feature selector.
        """
        return sorted(
            set(self._default_features(data_schema)) - set(self._ignore_features)
            if self._features is None
            else self._features
        )

    @abstractmethod
    def _fit(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, Any]:
        """Fits the feature selector to the specified DataFrame.

        Args:
            data_schema: DataSchema of the DataFrame.
            dataframe: DataFrame to be fitted.

        Raises:
            NotImplementedError: Must be implemented in the child class that inherits from `BaseFeatureSelector`.
        """
        raise NotImplementedError

    @abstractmethod
    def _transform(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, vaex.DataFrame]:
        """Transforms the dataframe by extracting the selected features.

        Args:
            data_schema: DataSchema of the DataFrame.
            dataframe: DataFrame to be transformed.

        Raises:
            NotImplementedError: Must be implemented in the child class that inherits from `BaseFeatureSelector`.
        """
        raise NotImplementedError

    @abstractmethod
    def _default_features(self, data_schema: DataSchema) -> tuple[str, ...]:
        """Returns the default set of features to be used by the feature selector.

        Args:
            data_schema: DataSchema of the DataFrame.

        Raises:
            NotImplementedError: Must be implemented in the child class that inherits from `BaseFeatureSelector`.
        """
        raise NotImplementedError

    @abstractmethod
    def _fingerprint(self) -> Hashable:
        """Returns a hashable object that uniquely identifies the feature selector.

        Note:
            Subclasses should call the parent method and include the result in the hashable object along with any other
            information that uniquely identifies the feature selector. All attributes that are used in the
            feature selector that affect the output of the feature selector should be included in the hashable object.

        Returns:
            Hashable object that uniquely identifies the feature selector.
        """
        return self.__class__.__name__, self._features, self._ignore_features



"""Module containing the TransformStep class."""

from __future__ import annotations

from typing import Union, cast

from typing_extensions import TypedDict
from vaex import DataFrame

from mleko.dataset.data_schema import DataSchema
from mleko.dataset.transform.base_transformer import BaseTransformer
from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import FitTransformAction, FitTransformPipelineStep
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr


logger = CustomLogger()
"""The logger for the module."""


class TransformStepInputType(TypedDict):
    """The input type of the TransformStep."""

    data_schema: Union[str, DataSchema]
    """DataSchema or the key for the DataSchema to be used for transformation."""

    dataframe: Union[str, DataFrame]
    """DataFrame or the key for the DataFrame to be transformed."""


class TransformStepOutputFitType(TypedDict):
    """The output type of the TransformStep when action is 'fit'."""

    data_schema: str
    """The key for the updated DataSchema after transformation."""

    transformer: str
    """The key for the fitted transformer after transformation."""


class TransformStepOutputTransformType(TypedDict):
    """The output type of the TransformStep when action is 'transform'."""

    data_schema: str
    """The key for the updated DataSchema after transformation."""

    dataframe: str
    """The key for the transformed DataFrame after transformation."""


class TransformStepOutputFitTransformType(TransformStepOutputFitType, TransformStepOutputTransformType):
    """The output type of the TransformStep when action is 'fit_transform'."""

    pass


class TransformStep(FitTransformPipelineStep):
    """Pipeline step for transformation of features in DataFrame."""

    _inputs: TransformStepInputType
    _outputs: TransformStepOutputFitType | TransformStepOutputTransformType | TransformStepOutputFitTransformType

    @auto_repr
    def __init__(
        self,
        transformer: BaseTransformer,
        action: FitTransformAction,
        inputs: TransformStepInputType,
        outputs: TransformStepOutputFitType | TransformStepOutputTransformType | TransformStepOutputFitTransformType,
        cache_group: str | None = None,
    ) -> None:
        """Initialize the TransformStep with the specified transformer.

        Args:
            transformer: The Transformer responsible for handling feature transformation.
            action: The action to perform, one of "fit", "transform", or "fit_transform".
            inputs: A dictionary of input keys following the `TransformStepInputType` schema.
            outputs: A dictionary of output keys following one of `TransformStepOutputFitType`,
                `TransformStepOutputTransformType`, or `TransformStepOutputFitTransformType` schemas, depending on
                the action.
            cache_group: The cache group to use.
        """
        super().__init__(action, inputs, outputs, cache_group)
        self._transformer = transformer

    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Perform transformation using the configured transformer.

        Args:
            data_container: Contains the input DataFrame.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Returns:
            A DataContainer containing the result depending on the action.
        """
        data_schema = self._validate_and_get_input(self._inputs["data_schema"], DataSchema, data_container)
        dataframe = self._validate_and_get_input(self._inputs["dataframe"], DataFrame, data_container)

        if self._action == "fit":
            self._outputs = cast(TransformStepOutputFitType, self._outputs)
            ds, transformer = self._transformer.fit(data_schema, dataframe, self._cache_group, force_recompute)
            data_container.data[self._outputs["data_schema"]] = ds
            data_container.data[self._outputs["transformer"]] = transformer
        elif self._action == "transform":
            self._outputs = cast(TransformStepOutputTransformType, self._outputs)
            ds, df = self._transformer.transform(data_schema, dataframe, self._cache_group, force_recompute)
            data_container.data[self._outputs["data_schema"]] = ds
            data_container.data[self._outputs["dataframe"]] = df
        elif self._action == "fit_transform":
            self._outputs = cast(TransformStepOutputFitTransformType, self._outputs)
            ds, transformer, df = self._transformer.fit_transform(
                data_schema, dataframe, self._cache_group, force_recompute
            )
            data_container.data[self._outputs["data_schema"]] = ds
            data_container.data[self._outputs["transformer"]] = transformer
            data_container.data[self._outputs["dataframe"]] = df
        return data_container

    def _get_input_model(self) -> type[TransformStepInputType]:
        """Get the input type for the TransformStep.

        Returns:
            Input type for the TransformStep.
        """
        return TransformStepInputType

    def _get_output_model(
        self,
    ) -> type[TransformStepOutputFitType | TransformStepOutputTransformType | TransformStepOutputFitTransformType]:
        """Get the output type for the TransformStep.

        Returns:
            Output type for the TransformStep.
        """
        if self._action == "fit":
            return TransformStepOutputFitType
        if self._action == "transform":
            return TransformStepOutputTransformType
        return TransformStepOutputFitTransformType



"""This module contains a LRU cache mixin that can be used by the cache classes.

The `LRUCacheMixin` can be used to add Least Recently Used (LRU) cache functionality to the cached classes.
It evicts the least recently used cache entries when the maximum number of cache entries is exceeded.
The LRU cache mechanism ensures that the most frequently accessed cache entries are retained, while entries that are
rarely accessed and have not been accessed recently are evicted first as the cache fills up. The cache entries
are stored in the cache directory, and the cache is trimmed if needed when cold starting the cache.
"""

from __future__ import annotations

import inspect
import re
from collections import OrderedDict, defaultdict
from pathlib import Path
from typing import Any, Sequence

from mleko.utils.custom_logger import CustomLogger

from .cache_mixin import CacheMixin, get_qualified_name_from_frame
from .handlers import CacheHandler


METHOD_GROUP_CACHE_KEY_PATTERN = r"\.([a-zA-Z_][a-zA-Z0-9_]*)(\.[a-zA-Z_][a-zA-Z0-9_]*)?\.[a-fA-F\d]{32}"
"""A regular expression pattern for matching cache keys in the format `.<method_name>.<cache_key>.<fingerprint>`."""

logger = CustomLogger()
"""A module-level custom logger."""


class LRUCacheMixin(CacheMixin):
    """Least Recently Used Cache Mixin.

    This mixin class extends the CacheMixin to provide a Least Recently Used (LRU) cache mechanism.
    It evicts the least recently used cache entries when the maximum number of cache entries is exceeded.
    The LRU cache mechanism ensures that the most frequently accessed cache entries are retained,
    while entries that are rarely accessed and have not been accessed recently are evicted first as the cache fills up.
    """

    def __init__(self, cache_directory: str | Path, cache_size: int) -> None:
        """Initializes the `LRUCacheMixin` with the provided cache directory and maximum number of cache entries.

        Note:
            The cache directory is created if it does not exist. When cold starting the cache, the cache will be loaded
            from the cache directory. The files are sorted by their modification time, and the cache is trimmed if
            needed.

        Args:
            cache_directory: The directory where cache files will be stored. If None, the cache will be disabled.
            cache_size: The maximum number of cache entries allowed before eviction.

        Examples:
            >>> from mleko.cache import LRUCacheMixin
            >>> class MyClass(LRUCacheMixin):
            ...     def __init__(self):
            ...         super().__init__("cache", "pkl", 2)
            ...
            ...     def my_method(self, x):
            ...         return self._cached_execute(lambda: x ** 2, [x])[1]
            >>> my_class = MyClass()
            >>> my_class.my_method(2)
            4 # This is not cached
            >>> my_class.my_method(2)
            4 # This is cached
            >>> my_class.my_method(3)
            9 # This is not cached
            >>> my_class.my_method(2)
            4 # This is cached
            >>> my_class.my_method(3)
            9 # This is cached
            >>> my_class.my_method(4)
            16 # This is not cached, and the cache is full so the least recently used entry is evicted (x = 2)
        """
        super().__init__(cache_directory, cache_size <= 0)
        if not self._disable_cache:
            self._cache_size = cache_size
            self._cache: dict[str, OrderedDict[str, bool]] = defaultdict(OrderedDict)
            self._load_cache_from_disk()

    def _load_cache_from_disk(self) -> None:
        """Loads the cache entries from the cache directory and initializes the LRU cache.

        Cache entries are ordered by their modification time, and the cache is trimmed if needed.
        """
        frame_qualname = get_qualified_name_from_frame(inspect.stack()[2])
        class_name = frame_qualname.split(".")[-2]
        file_name_pattern = rf"{class_name}{METHOD_GROUP_CACHE_KEY_PATTERN}"

        cache_files = [f for f in self._cache_directory.glob("*") if re.search(file_name_pattern, str(f.stem))]
        ordered_cache_files = list(
            dict.fromkeys([f.stem for f in sorted(cache_files, key=lambda x: x.stat().st_mtime)])
        )

        for cache_file in ordered_cache_files:
            cache_key_match = re.search(file_name_pattern, cache_file)
            if cache_key_match:
                method_name, cache_group = cache_key_match.groups()
                group_identifier = (
                    f"{class_name}.{method_name}{cache_group}" if cache_group else f"{class_name}.{method_name}"
                )
                cache_key = cache_key_match.group(0)

                if cache_key not in self._cache[group_identifier]:
                    self._evict_least_recently_used_if_full(group_identifier)
                    self._cache[group_identifier][cache_key] = True

    def _load_from_cache(
        self,
        cache_key: str,
        cache_handlers: CacheHandler | list[CacheHandler],
    ) -> Any | None:
        """Loads data from the cache based on the provided cache key and updates the LRU cache.

        Args:
            cache_key: A string representing the cache key.
            cache_handlers: A CacheHandler instance or a list of CacheHandler instances. If a single CacheHandler
                instance is provided, it will be used for all cache files. If a list of CacheHandler instances is
                provided, each CacheHandler instance will be used for each cache file.

        Returns:
            The cached data if it exists, or None if there is no data for the given cache key.
        """
        for group_identifier in self._cache.keys():
            if cache_key in self._cache[group_identifier]:
                self._cache[group_identifier].move_to_end(cache_key)
                return super()._load_from_cache(cache_key, cache_handlers)
        return None

    def _save_to_cache(
        self,
        cache_key: str,
        output: Any | Sequence[Any],
        cache_handlers: CacheHandler | list[CacheHandler],
    ) -> None:
        """Saves the given data to the cache using the provided cache key, updating the LRU cache accordingly.

        If the cache reaches its maximum size, the least recently used entry will be evicted.

        Args:
            cache_key: A string representing the cache key.
            output: The data to be saved to the cache.
            cache_handlers: A CacheHandler instance or a list of CacheHandler instances. If a single CacheHandler
                instance is provided, it will be used for all cache files. If a list of CacheHandler instances is
                provided, each CacheHandler instance will be used for each cache file.
        """
        frame_qualname = get_qualified_name_from_frame(inspect.stack()[2])
        class_name = frame_qualname.split(".")[-2]
        cache_key_match = re.match(rf"[a-zA-Z_][a-zA-Z0-9_]*{METHOD_GROUP_CACHE_KEY_PATTERN}", cache_key)

        if cache_key_match:
            method_name, cache_group = cache_key_match.groups()
            group_identifier = (
                f"{class_name}.{method_name}{cache_group}" if cache_group else f"{class_name}.{method_name}"
            )

            if cache_key not in self._cache[group_identifier]:
                self._evict_least_recently_used_if_full(group_identifier)
                self._cache[group_identifier][cache_key] = True
            else:
                self._cache[group_identifier].move_to_end(cache_key)

            super()._save_to_cache(cache_key, output, cache_handlers)

    def _evict_least_recently_used_if_full(self, group_identifier: str) -> None:
        """Evicts the least recently used cache entry if the cache is full.

        Args:
            group_identifier: The group identifier for the cache entries.
        """
        if len(self._cache[group_identifier]) >= self._cache_size:
            oldest_key = next(iter(self._cache[group_identifier]))
            del self._cache[group_identifier][oldest_key]
            for file in self._cache_directory.glob(f"{oldest_key}*.*"):
                logger.debug(f"Max cache size reached ({self._cache_size}), deleting file {file}")
                file.unlink()
